# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import namedtuple
from textwrap import indent

from numba.cuda.types import float32, float64, int16, int32, int64, void, Tuple
from numba.cuda.typing.templates import signature

arg = namedtuple("arg", ("name", "ty", "is_ptr"))

functions = {
    "__nv_abs": (int32, [arg(name="x", ty=int32, is_ptr=False)]),
    "__nv_acos": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_acosf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_acosh": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_acoshf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_asin": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_asinf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_asinh": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_asinhf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_atan": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_atan2": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_atan2f": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_atanf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_atanh": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_atanhf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_brev": (int32, [arg(name="x", ty=int32, is_ptr=False)]),
    "__nv_brevll": (int64, [arg(name="x", ty=int64, is_ptr=False)]),
    "__nv_byte_perm": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
            arg(name="z", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_cbrt": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_cbrtf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_ceil": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_ceilf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_clz": (int32, [arg(name="x", ty=int32, is_ptr=False)]),
    "__nv_clzll": (int32, [arg(name="x", ty=int64, is_ptr=False)]),
    "__nv_copysign": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_copysignf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_cos": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_cosf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_cosh": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_coshf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_cospi": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_cospif": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_dadd_rd": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_dadd_rn": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_dadd_ru": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_dadd_rz": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_ddiv_rd": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_ddiv_rn": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_ddiv_ru": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_ddiv_rz": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_dmul_rd": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_dmul_rn": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_dmul_ru": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_dmul_rz": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_double2float_rd": (
        float32,
        [arg(name="d", ty=float64, is_ptr=False)],
    ),
    "__nv_double2float_rn": (
        float32,
        [arg(name="d", ty=float64, is_ptr=False)],
    ),
    "__nv_double2float_ru": (
        float32,
        [arg(name="d", ty=float64, is_ptr=False)],
    ),
    "__nv_double2float_rz": (
        float32,
        [arg(name="d", ty=float64, is_ptr=False)],
    ),
    "__nv_double2hiint": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2int_rd": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2int_rn": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2int_ru": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2int_rz": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2ll_rd": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double2ll_rn": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double2ll_ru": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double2ll_rz": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double2loint": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2uint_rd": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2uint_rn": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2uint_ru": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2uint_rz": (int32, [arg(name="d", ty=float64, is_ptr=False)]),
    "__nv_double2ull_rd": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double2ull_rn": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double2ull_ru": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double2ull_rz": (int64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_double_as_longlong": (
        int64,
        [arg(name="x", ty=float64, is_ptr=False)],
    ),
    "__nv_drcp_rd": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_drcp_rn": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_drcp_ru": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_drcp_rz": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_dsqrt_rd": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_dsqrt_rn": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_dsqrt_ru": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_dsqrt_rz": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_erf": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_erfc": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_erfcf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_erfcinv": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_erfcinvf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_erfcx": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_erfcxf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_erff": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_erfinv": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_erfinvf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_exp": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_exp10": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_exp10f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_exp2": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_exp2f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_expf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_expm1": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_expm1f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fabs": (float64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_fabsf": (float32, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_fadd_rd": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fadd_rn": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fadd_ru": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fadd_rz": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fast_cosf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fast_exp10f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fast_expf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fast_fdividef": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fast_log10f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fast_log2f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fast_logf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fast_powf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fast_sincosf": (
        void,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="sptr", ty=float32, is_ptr=True),
            arg(name="cptr", ty=float32, is_ptr=True),
        ],
    ),
    "__nv_fast_sinf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fast_tanf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fdim": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fdimf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fdiv_rd": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fdiv_rn": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fdiv_ru": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fdiv_rz": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_ffs": (int32, [arg(name="x", ty=int32, is_ptr=False)]),
    "__nv_ffsll": (int32, [arg(name="x", ty=int64, is_ptr=False)]),
    "__nv_finitef": (int32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_float2half_rn": (int16, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2int_rd": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2int_rn": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2int_ru": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2int_rz": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2ll_rd": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2ll_rn": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2ll_ru": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2ll_rz": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2uint_rd": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2uint_rn": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2uint_ru": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2uint_rz": (int32, [arg(name="in", ty=float32, is_ptr=False)]),
    "__nv_float2ull_rd": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2ull_rn": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2ull_ru": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float2ull_rz": (int64, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_float_as_int": (int32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_floor": (float64, [arg(name="f", ty=float64, is_ptr=False)]),
    "__nv_floorf": (float32, [arg(name="f", ty=float32, is_ptr=False)]),
    "__nv_fma": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
            arg(name="z", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fma_rd": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
            arg(name="z", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fma_rn": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
            arg(name="z", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fma_ru": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
            arg(name="z", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fma_rz": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
            arg(name="z", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fmaf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
            arg(name="z", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmaf_rd": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
            arg(name="z", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmaf_rn": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
            arg(name="z", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmaf_ru": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
            arg(name="z", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmaf_rz": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
            arg(name="z", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmax": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fmaxf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmin": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fminf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmod": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_fmodf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmul_rd": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmul_rn": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmul_ru": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fmul_rz": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_frcp_rd": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_frcp_rn": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_frcp_ru": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_frcp_rz": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_frexp": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="b", ty=int32, is_ptr=True),
        ],
    ),
    "__nv_frexpf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="b", ty=int32, is_ptr=True),
        ],
    ),
    "__nv_frsqrt_rn": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fsqrt_rd": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fsqrt_rn": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fsqrt_ru": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fsqrt_rz": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_fsub_rd": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fsub_rn": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fsub_ru": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_fsub_rz": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_hadd": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_half2float": (float32, [arg(name="h", ty=int16, is_ptr=False)]),
    "__nv_hiloint2double": (
        float64,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_hypot": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_hypotf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_ilogb": (int32, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_ilogbf": (int32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_int2double_rn": (float64, [arg(name="i", ty=int32, is_ptr=False)]),
    "__nv_int2float_rd": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_int2float_rn": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_int2float_ru": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_int2float_rz": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_int_as_float": (float32, [arg(name="x", ty=int32, is_ptr=False)]),
    "__nv_isfinited": (int32, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_isinfd": (int32, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_isinff": (int32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_isnand": (int32, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_isnanf": (int32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_j0": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_j0f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_j1": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_j1f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_jn": (
        float64,
        [
            arg(name="n", ty=int32, is_ptr=False),
            arg(name="x", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_jnf": (
        float32,
        [
            arg(name="n", ty=int32, is_ptr=False),
            arg(name="x", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_ldexp": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_ldexpf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_lgamma": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_lgammaf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_ll2double_rd": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ll2double_rn": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ll2double_ru": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ll2double_rz": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ll2float_rd": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ll2float_rn": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ll2float_ru": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ll2float_rz": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_llabs": (int64, [arg(name="x", ty=int64, is_ptr=False)]),
    "__nv_llmax": (
        int64,
        [
            arg(name="x", ty=int64, is_ptr=False),
            arg(name="y", ty=int64, is_ptr=False),
        ],
    ),
    "__nv_llmin": (
        int64,
        [
            arg(name="x", ty=int64, is_ptr=False),
            arg(name="y", ty=int64, is_ptr=False),
        ],
    ),
    "__nv_llrint": (int64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_llrintf": (int64, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_llround": (int64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_llroundf": (int64, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_log": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_log10": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_log10f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_log1p": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_log1pf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_log2": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_log2f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_logb": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_logbf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_logf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_longlong_as_double": (
        float64,
        [arg(name="x", ty=int64, is_ptr=False)],
    ),
    "__nv_max": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_min": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_modf": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="b", ty=float64, is_ptr=True),
        ],
    ),
    "__nv_modff": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="b", ty=float32, is_ptr=True),
        ],
    ),
    "__nv_mul24": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_mul64hi": (
        int64,
        [
            arg(name="x", ty=int64, is_ptr=False),
            arg(name="y", ty=int64, is_ptr=False),
        ],
    ),
    "__nv_mulhi": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    # __nv_nan and __nv_nanf are excluded - they return a representation of a
    # quiet NaN, but the argument they take seems to be undocumented, and
    # follows a strange form - it is not an output like every other pointer
    # argument. If a NaN is required, one can be obtained in CUDA Python by
    # other means, e.g. `math.nan`. They are left in this list for completeness
    # / reference.
    # "__nv_nan": (float64, [arg(name="tagp", ty=int8, is_ptr=True)]),
    # "__nv_nanf": (float32, [arg(name="tagp", ty=int8, is_ptr=True)]),
    "__nv_nearbyint": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_nearbyintf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_nextafter": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_nextafterf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_normcdf": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_normcdff": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_normcdfinv": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_normcdfinvf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_popc": (int32, [arg(name="x", ty=int32, is_ptr=False)]),
    "__nv_popcll": (int32, [arg(name="x", ty=int64, is_ptr=False)]),
    "__nv_pow": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_powf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_powi": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_powif": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_rcbrt": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_rcbrtf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_remainder": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_remainderf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
        ],
    ),
    "__nv_remquo": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=float64, is_ptr=False),
            arg(name="c", ty=int32, is_ptr=True),
        ],
    ),
    "__nv_remquof": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=float32, is_ptr=False),
            arg(name="quo", ty=int32, is_ptr=True),
        ],
    ),
    "__nv_rhadd": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_rint": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_rintf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_round": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_roundf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_rsqrt": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_rsqrtf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_sad": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
            arg(name="z", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_saturatef": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_scalbn": (
        float64,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_scalbnf": (
        float32,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_signbitd": (int32, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_signbitf": (int32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_sin": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_sincos": (
        void,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="sptr", ty=float64, is_ptr=True),
            arg(name="cptr", ty=float64, is_ptr=True),
        ],
    ),
    "__nv_sincosf": (
        void,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="sptr", ty=float32, is_ptr=True),
            arg(name="cptr", ty=float32, is_ptr=True),
        ],
    ),
    "__nv_sincospi": (
        void,
        [
            arg(name="x", ty=float64, is_ptr=False),
            arg(name="sptr", ty=float64, is_ptr=True),
            arg(name="cptr", ty=float64, is_ptr=True),
        ],
    ),
    "__nv_sincospif": (
        void,
        [
            arg(name="x", ty=float32, is_ptr=False),
            arg(name="sptr", ty=float32, is_ptr=True),
            arg(name="cptr", ty=float32, is_ptr=True),
        ],
    ),
    "__nv_sinf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_sinh": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_sinhf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_sinpi": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_sinpif": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_sqrt": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_sqrtf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_tan": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_tanf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_tanh": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_tanhf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_tgamma": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_tgammaf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_trunc": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_truncf": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_uhadd": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_uint2double_rn": (float64, [arg(name="i", ty=int32, is_ptr=False)]),
    "__nv_uint2float_rd": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_uint2float_rn": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_uint2float_ru": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_uint2float_rz": (float32, [arg(name="in", ty=int32, is_ptr=False)]),
    "__nv_ull2double_rd": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ull2double_rn": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ull2double_ru": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ull2double_rz": (float64, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ull2float_rd": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ull2float_rn": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ull2float_ru": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ull2float_rz": (float32, [arg(name="l", ty=int64, is_ptr=False)]),
    "__nv_ullmax": (
        int64,
        [
            arg(name="x", ty=int64, is_ptr=False),
            arg(name="y", ty=int64, is_ptr=False),
        ],
    ),
    "__nv_ullmin": (
        int64,
        [
            arg(name="x", ty=int64, is_ptr=False),
            arg(name="y", ty=int64, is_ptr=False),
        ],
    ),
    "__nv_umax": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_umin": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_umul24": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_umul64hi": (
        int64,
        [
            arg(name="x", ty=int64, is_ptr=False),
            arg(name="y", ty=int64, is_ptr=False),
        ],
    ),
    "__nv_umulhi": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_urhadd": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_usad": (
        int32,
        [
            arg(name="x", ty=int32, is_ptr=False),
            arg(name="y", ty=int32, is_ptr=False),
            arg(name="z", ty=int32, is_ptr=False),
        ],
    ),
    "__nv_y0": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_y0f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_y1": (float64, [arg(name="x", ty=float64, is_ptr=False)]),
    "__nv_y1f": (float32, [arg(name="x", ty=float32, is_ptr=False)]),
    "__nv_yn": (
        float64,
        [
            arg(name="n", ty=int32, is_ptr=False),
            arg(name="x", ty=float64, is_ptr=False),
        ],
    ),
    "__nv_ynf": (
        float32,
        [
            arg(name="n", ty=int32, is_ptr=False),
            arg(name="x", ty=float32, is_ptr=False),
        ],
    ),
}


def create_signature(retty, args):
    """
    Given the return type and arguments for a libdevice function, return the
    signature of the stub function used to call it from CUDA Python.
    """

    # Any pointer arguments should be part of the return type.
    return_types = [arg.ty for arg in args if arg.is_ptr]
    # If the return type is void, there is no point adding it to the list of
    # return types.
    if retty != void:
        return_types.insert(0, retty)

    if len(return_types) > 1:
        retty = Tuple(return_types)
    else:
        retty = return_types[0]

    argtypes = [arg.ty for arg in args if not arg.is_ptr]

    return signature(retty, *argtypes)


# The following code generates the stubs for libdevice functions.
#
# Stubs can be regenerated (e.g. if the functions dict above is modified) with:
#
# python -c "from numba.cuda.libdevicefuncs import generate_stubs; \
#            generate_stubs()" > numba/cuda/libdevice.py

docstring_template = """
See https://docs.nvidia.com/cuda/libdevice-users-guide/{func}.html

{param_types}
:rtype: {retty}
"""

param_template = """\
:param {a.name}: Argument.
:type {a.name}: {a.ty}"""


def generate_stubs():
    for name, (retty, args) in functions.items():
        # Some libdevice functions have arguments called `in`, which causes a
        # syntax error in Python, so we rename these to `x`.
        def argname(arg):
            if arg.name == "in":
                return "x"
            else:
                return arg.name

        argnames = [argname(a) for a in args if not a.is_ptr]
        argstr = ", ".join(argnames)
        signature = create_signature(retty, args)

        param_types = "\n".join(
            [param_template.format(a=a) for a in args if not a.is_ptr]
        )
        docstring = docstring_template.format(
            param_types=param_types, retty=signature.return_type, func=name
        )
        docstring = indent(docstring, "    ")
        print(f'def {name[5:]}({argstr}):\n    """{docstring}"""\n\n')



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import namedtuple
from functools import singledispatch
import ctypes
import enum
import functools
import operator

import numpy as np
from numpy.random.bit_generator import BitGenerator
from numba.cuda import HAS_NUMBA
from numba.cuda import types
from numba.cuda.core import errors
from numba.cuda import utils
from numba.cuda.np import numpy_support

from cuda.core import Buffer
from cuda.core.utils import StridedMemoryView


# terminal color markup
_termcolor = errors.termcolor()


class Purpose(enum.Enum):
    # Value being typed is used as an argument
    argument = 1
    # Value being typed is used as a constant
    constant = 2


_TypeofContext = namedtuple("_TypeofContext", ("purpose",))


def typeof(val, purpose=Purpose.argument):
    """
    Get the Numba type of a Python value for the given purpose.
    """
    # Note the behaviour for Purpose.argument must match _typeof.c.
    c = _TypeofContext(purpose)
    ty = typeof_impl(val, c)
    if ty is None:
        msg = _termcolor.errmsg(f"Cannot determine Numba type of {type(val)}")
        raise ValueError(msg)
    return ty


@singledispatch
def typeof_impl(val, c):
    """
    Generic typeof() implementation.
    """
    tp = getattr(val, "_numba_type_", None)
    if tp is not None:
        return tp

    # Check for dlpack objects
    dlpack = getattr(val, "__dlpack__", None)
    if dlpack is not None:
        tp = _typeof_dlpack(dlpack, c)
        if tp is not None:
            return tp

    # Check for __cuda_array_interface__ objects (third-party device arrays)

    # Numba's own DeviceNDArray is handled above via _numba_type_.
    cai = getattr(val, "__cuda_array_interface__", None)
    if cai is not None:
        tp = _typeof_cuda_array_interface(cai, c)
        if tp is not None:
            return tp

    tp = _typeof_buffer(val, c)
    if tp is not None:
        return tp

    # cffi is handled here as it does not expose a public base class
    # for exported functions or CompiledFFI instances.
    from numba.cuda.typing import cffi_utils

    if cffi_utils.SUPPORTED:
        if cffi_utils.is_cffi_func(val):
            return cffi_utils.make_function_type(val)
        if cffi_utils.is_ffi_instance(val):
            return types.ffi

    if HAS_NUMBA:
        # Fallback to Numba's typeof_impl for third-party registrations
        from numba.core.typing.typeof import typeof_impl as core_typeof_impl

        tp = core_typeof_impl(val, c)
        if tp is not None:
            return tp

    return None


def _typeof_buffer(val, c):
    from numba.cuda.typing import bufproto

    try:
        m = memoryview(val)
    except TypeError:
        return
    # Object has the buffer protocol
    try:
        dtype = bufproto.decode_pep3118_format(m.format, m.itemsize)
    except ValueError:
        return
    type_class = bufproto.get_type_class(type(val))
    layout = bufproto.infer_layout(m)
    return type_class(dtype, m.ndim, layout=layout, readonly=m.readonly)


@typeof_impl.register(ctypes._CFuncPtr)
def _typeof_ctypes_function(val, c):
    from .ctypes_utils import is_ctypes_funcptr, make_function_type

    if is_ctypes_funcptr(val):
        return make_function_type(val)


@typeof_impl.register(type)
def _typeof_type(val, c):
    """
    Type various specific Python types.
    """
    if issubclass(val, BaseException):
        return types.ExceptionClass(val)
    if issubclass(val, tuple) and hasattr(val, "_asdict"):
        return types.NamedTupleClass(val)

    if issubclass(val, np.generic):
        return types.NumberClass(numpy_support.from_dtype(val))

    if issubclass(val, types.Type):
        return types.TypeRef(val)


@typeof_impl.register(bool)
def _typeof_bool(val, c):
    return types.boolean


@typeof_impl.register(float)
def _typeof_float(val, c):
    return types.float64


@typeof_impl.register(complex)
def _typeof_complex(val, c):
    return types.complex128


@typeof_impl.register(int)
def _typeof_int(val, c):
    # As in _typeof.c
    nbits = utils.bit_length(val)
    if nbits < 32:
        typ = types.intp
    elif nbits < 64:
        typ = types.int64
    elif nbits == 64 and val >= 0:
        typ = types.uint64
    else:
        raise ValueError("Int value is too large: %s" % val)
    return typ


@typeof_impl.register(np.generic)
def _typeof_numpy_scalar(val, c):
    try:
        return numpy_support.map_arrayscalar_type(val)
    except errors.NumbaNotImplementedError:
        pass
    except NotImplementedError:
        pass


@typeof_impl.register(str)
def _typeof_str(val, c):
    return types.string


@typeof_impl.register(type((lambda a: a).__code__))
def _typeof_code(val, c):
    return types.code_type


@typeof_impl.register(type(None))
def _typeof_none(val, c):
    return types.none


@typeof_impl.register(type(Ellipsis))
def _typeof_ellipsis(val, c):
    return types.ellipsis


@typeof_impl.register(tuple)
def _typeof_tuple(val, c):
    tys = [typeof_impl(v, c) for v in val]
    if any(ty is None for ty in tys):
        return
    return types.BaseTuple.from_types(tys, type(val))


@typeof_impl.register(list)
def _typeof_list(val, c):
    if len(val) == 0:
        raise ValueError("Cannot type empty list")
    ty = typeof_impl(val[0], c)
    if ty is None:
        raise ValueError(f"Cannot type list element type {type(val[0])}")
    return types.List(ty, reflected=True)


@typeof_impl.register(set)
def _typeof_set(val, c):
    if len(val) == 0:
        raise ValueError("Cannot type empty set")
    item = next(iter(val))
    ty = typeof_impl(item, c)
    if ty is None:
        raise ValueError(f"Cannot type set element type {type(item)}")
    return types.Set(ty, reflected=True)


@typeof_impl.register(slice)
def _typeof_slice(val, c):
    return types.slice2_type if val.step in (None, 1) else types.slice3_type


@typeof_impl.register(enum.Enum)
@typeof_impl.register(enum.IntEnum)
def _typeof_enum(val, c):
    clsty = typeof_impl(type(val), c)
    return clsty.member_type


@typeof_impl.register(enum.EnumMeta)
def _typeof_enum_class(val, c):
    cls = val
    members = list(cls.__members__.values())
    if len(members) == 0:
        raise ValueError("Cannot type enum with no members")
    dtypes = {typeof_impl(mem.value, c) for mem in members}
    if len(dtypes) > 1:
        raise ValueError(
            "Cannot type heterogeneous enum: "
            "got value types %s" % ", ".join(sorted(str(ty) for ty in dtypes))
        )
    if issubclass(val, enum.IntEnum):
        typecls = types.IntEnumClass
    else:
        typecls = types.EnumClass
    return typecls(cls, dtypes.pop())


@typeof_impl.register(np.dtype)
def _typeof_dtype(val, c):
    tp = numpy_support.from_dtype(val)
    return types.DType(tp)


@typeof_impl.register(np.ndarray)
def _typeof_ndarray(val, c):
    if isinstance(val, np.ma.MaskedArray):
        msg = "Unsupported array type: numpy.ma.MaskedArray."
        raise errors.NumbaTypeError(msg)
    try:
        dtype = numpy_support.from_dtype(val.dtype)
    except errors.NumbaNotImplementedError:
        raise errors.NumbaValueError(f"Unsupported array dtype: {val.dtype}")
    layout = numpy_support.map_layout(val)
    readonly = not val.flags.writeable
    return types.Array(dtype, val.ndim, layout, readonly=readonly)


@typeof_impl.register(types.NumberClass)
def _typeof_number_class(val, c):
    return val


@typeof_impl.register(types.Literal)
def _typeof_literal(val, c):
    return val


@typeof_impl.register(types.TypeRef)
def _typeof_typeref(val, c):
    return val


@typeof_impl.register(types.Type)
def _typeof_nb_type(val, c):
    if isinstance(val, types.BaseFunction):
        return val
    elif isinstance(val, (types.Number, types.Boolean)):
        return types.NumberClass(val)
    else:
        return types.TypeRef(val)


@typeof_impl.register(BitGenerator)
def typeof_numpy_random_bitgen(val, c):
    return types.NumPyRandomBitGeneratorType(val)


@typeof_impl.register(np.random.Generator)
def typeof_random_generator(val, c):
    return types.NumPyRandomGeneratorType(val)


@typeof_impl.register(np.polynomial.polynomial.Polynomial)
def typeof_numpy_polynomial(val, c):
    coef = typeof(val.coef)
    domain = typeof(val.domain)
    window = typeof(val.window)
    return types.PolynomialType(coef, domain, window)


@functools.lru_cache
def _typeof_cuda_array_interface_cached(
    *, dtype, shape, strides, readonly: bool
):
    # Determine layout
    ndim = len(shape)
    if not ndim:
        layout = "C"
    elif strides is None:
        layout = "C"
    else:
        itemsize = dtype.bitwidth
        # Quick rejection: C-contiguous has strides[-1] == itemsize,
        # F-contiguous has strides[0] == itemsize. If neither, it's "A".
        if strides[-1] == itemsize:
            c_strides = numpy_support.strides_from_shape(
                shape=shape,
                itemsize=itemsize,
                c_contiguous=True,
                f_contiguous=False,
            )
            layout = "C" if all(map(operator.eq, strides, c_strides)) else "A"
        elif strides[0] == itemsize:
            f_strides = numpy_support.strides_from_shape(
                shape=shape,
                itemsize=itemsize,
                c_contiguous=False,
                f_contiguous=True,
            )
            layout = "F" if all(map(operator.eq, strides, f_strides)) else "A"
        else:
            layout = "A"

    return types.Array(dtype, ndim, layout, readonly=readonly)


def _typeof_dlpack(val, c):
    obj = getattr(val, "__self__", None)
    if obj is not None:
        try:
            smv = StridedMemoryView.from_dlpack(obj, stream_ptr=-1)
        except BufferError:
            return

        smv_layout = smv._layout
        layout = (
            "C"
            if smv_layout.is_contiguous_c
            else "F"
            if smv_layout.is_contiguous_f
            else "A"
        )
        return types.Array(
            dtype=numpy_support.from_dtype(smv.dtype),
            ndim=len(smv.shape),
            layout=layout,
            readonly=smv.readonly,
        )


@functools.lru_cache
def _numba_dtype_from_str(typestr):
    return numpy_support.from_dtype(np.dtype(typestr))


def _typeof_cuda_array_interface(val, c):
    """
    Determine the type of a __cuda_array_interface__ object.

    This handles third-party device arrays that implement the CUDA
    Array Interface. These are typed as regular Array types, with lowering
    handled in numba.cuda.np.arrayobj.
    """
    dtype = _numba_dtype_from_str(val["typestr"])
    shape = val["shape"]
    strides = val.get("strides")
    _, readonly = val["data"]
    return _typeof_cuda_array_interface_cached(
        dtype=dtype,
        shape=shape,
        strides=strides,
        readonly=readonly,
    )


@typeof_impl.register(Buffer)
def typeof_buffer(val, c):
    return types.Array(
        dtype=_numba_dtype_from_str("uint8"),
        ndim=1,
        layout="C",
        readonly=False,
    )


@typeof_impl.register(StridedMemoryView)
def typeof_strided_memory_view(val, c):
    raw_layout = val._layout
    if raw_layout.is_contiguous_c:
        layout = "C"
    elif raw_layout.is_contiguous_f:
        layout = "F"
    elif raw_layout.is_contiguous_any:
        layout = "A"
    else:
        raise ValueError(
            "Unsupported StridedMemoryView layout; must be contiguous"
        )
    return types.Array(
        dtype=numpy_support.from_dtype(val.dtype),
        ndim=len(val.shape),
        layout=layout,
        readonly=val.readonly,
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

# delvewheel: patch

import importlib
from numba.cuda.core import config
from .utils import _readenv
import warnings
import sys


# Re-export types itself
import numba.cuda.types as types

# Re-export all type names
from numba.cuda.types import *

HAS_NUMBA = importlib.util.find_spec("numba") is not None

# Require NVIDIA CUDA bindings at import time
if not (
    importlib.util.find_spec("cuda")
    and importlib.util.find_spec("cuda.bindings")
):
    raise ImportError(
        "NVIDIA CUDA Python bindings not found. Install the 'cuda' package "
        '(e.g. pip install "cuda-bindings==XY.*" or "numba-cuda[cuXY]", '
        "with XY=12 or XY=13)."
    )

if config.ENABLE_CUDASIM:
    from .simulator_init import *
else:
    from .device_init import *
    from .device_init import _auto_device

from numba.cuda.compiler import (
    compile,
    compile_for_current_device,
    compile_ptx,
    compile_ptx_for_current_device,
    compile_all,
)

# This is the out-of-tree NVIDIA-maintained target. This is reported in Numba
# sysinfo (`numba -s`):
implementation = "NVIDIA"


# The default compute capability as set by the upstream Numba implementation.
config_default_cc = config.CUDA_DEFAULT_PTX_CC

# The default compute capability for Numba-CUDA. This will usually override the
# upstream Numba built-in default of 5.0, unless the user has set it even
# higher, in which case we should use the user-specified value. This default is
# aligned with recent toolkit versions.
numba_cuda_default_ptx_cc = (7, 5)

if numba_cuda_default_ptx_cc > config_default_cc:
    config.CUDA_DEFAULT_PTX_CC = numba_cuda_default_ptx_cc


# Warn if on Linux and RTLD_GLOBAL is enabled
if sys.platform.startswith("linux") and (sys.getdlopenflags() & 0x100) != 0:
    warnings.warn(
        "RTLD_GLOBAL is enabled, which might result in symbol resolution "
        "conflicts when importing both numba and numba.cuda. Consider using "
        "sys.setdlopenflags() to disable RTLD_GLOBAL "
        "if you encounter symbol conflicts."
    )

from numba.cuda.np.ufunc import vectorize, guvectorize



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import sys
import numba.cuda.types as types
from numba.cuda._internal.cuda_fp16 import (
    typing_registry,
    target_registry,
    __half,
    __double2half,
    __float2half,
    __float2half_rd,
    __float2half_rn,
    __float2half_ru,
    __float2half_rz,
    __int2half_rd,
    __int2half_rn,
    __int2half_ru,
    __int2half_rz,
    __ll2half_rd,
    __ll2half_rn,
    __ll2half_ru,
    __ll2half_rz,
    __short2half_rd,
    __short2half_rn,
    __short2half_ru,
    __short2half_rz,
    __uint2half_rd,
    __uint2half_rn,
    __uint2half_ru,
    __uint2half_rz,
    __ull2half_rd,
    __ull2half_rn,
    __ull2half_ru,
    __ull2half_rz,
    __ushort2half_rd,
    __ushort2half_rn,
    __ushort2half_ru,
    __ushort2half_rz,
    __half2char_rz,
    __half2float,
    __half2int_rd,
    __half2int_rn,
    __half2int_ru,
    __half2int_rz,
    __half2ll_rd,
    __half2ll_rn,
    __half2ll_ru,
    __half2ll_rz,
    __half2short_rd,
    __half2short_rn,
    __half2short_ru,
    __half2short_rz,
    __half2uchar_rz,
    __half2uint_rd,
    __half2uint_rn,
    __half2uint_ru,
    __half2uint_rz,
    __half2ull_rd,
    __half2ull_rn,
    __half2ull_ru,
    __half2ull_rz,
    __half2ushort_rd,
    __half2ushort_rn,
    __half2ushort_ru,
    __half2ushort_rz,
    __short_as_half,
    __ushort_as_half,
    __half_as_short,
    __half_as_ushort,
    __habs as habs,
    __habs,
    __hadd as hadd,
    __hadd,
    __hadd_rn,
    __hadd_sat,
    __hdiv as hdiv,
    __hdiv,
    __heq as heq,
    __heq,
    __hequ,
    __hfma as hfma,
    __hfma,
    __hfma_relu,
    __hfma_sat,
    __hge as hge,
    __hge,
    __hgeu,
    __hgt as hgt,
    __hgt,
    __hgtu,
    __hisinf,
    __hisnan,
    __hle as hle,
    __hle,
    __hleu,
    __hlt as hlt,
    __hlt,
    __hltu,
    __hmax as hmax,
    __hmax,
    __hmax_nan,
    __hmin as hmin,
    __hmin,
    __hmin_nan,
    __hmul as hmul,
    __hmul,
    __hmul_rn,
    __hmul_sat,
    __hne as hne,
    __hne,
    __hneg as hneg,
    __hneg,
    __hneu,
    __hsub as hsub,
    __hsub,
    __hsub_rn,
    __hsub_sat,
    atomicAdd,
    hceil,
    hcos,
    hexp,
    hexp10,
    hexp2,
    hfloor,
    hlog,
    hlog10,
    hlog2,
    hrcp,
    hrint,
    hrsqrt,
    hsin,
    hsqrt,
    htanh,
    htanh_approx,
    htrunc,
)

from numba.cuda.extending import overload
import math


def _make_unary(a, func):
    if isinstance(a, types.Float) and a.bitwidth == 16:
        return lambda a: func(a)


# Bind low++ bindings to math APIs
@overload(math.trunc, target="cuda")
def trunc_ol(a):
    return _make_unary(a, htrunc)


@overload(math.ceil, target="cuda")
def ceil_ol(a):
    return _make_unary(a, hceil)


@overload(math.floor, target="cuda")
def floor_ol(a):
    return _make_unary(a, hfloor)


@overload(math.fabs, target="cuda")
def fabs_ol(a):
    return _make_unary(a, habs)


@overload(math.sqrt, target="cuda")
def sqrt_ol(a):
    return _make_unary(a, hsqrt)


@overload(math.log, target="cuda")
def log_ol(a):
    return _make_unary(a, hlog)


@overload(math.log2, target="cuda")
def log2_ol(a):
    return _make_unary(a, hlog2)


@overload(math.log10, target="cuda")
def log10_ol(a):
    return _make_unary(a, hlog10)


@overload(math.exp, target="cuda")
def exp_ol(a):
    return _make_unary(a, hexp)


if sys.version_info >= (3, 11):

    @overload(math.exp2, target="cuda")
    def exp2_ol(a):
        return _make_unary(a, hexp2)


@overload(math.tanh, target="cuda")
def tanh_ol(a):
    return _make_unary(a, htanh)


@overload(math.cos, target="cuda")
def cos_ol(a):
    return _make_unary(a, hcos)


@overload(math.sin, target="cuda")
def sin_ol(a):
    return _make_unary(a, hsin)


try:
    from math import exp2

    @overload(exp2, target="cuda")
    def exp2_ol(a):
        return _make_unary(a, hexp2)
except ImportError:
    pass


__all__ = [
    "typing_registry",
    "target_registry",
    "__half",
    "__double2half",
    "__float2half",
    "__float2half_rd",
    "__float2half_rn",
    "__float2half_ru",
    "__float2half_rz",
    "__int2half_rd",
    "__int2half_rn",
    "__int2half_ru",
    "__int2half_rz",
    "__ll2half_rd",
    "__ll2half_rn",
    "__ll2half_ru",
    "__ll2half_rz",
    "__short2half_rd",
    "__short2half_rn",
    "__short2half_ru",
    "__short2half_rz",
    "__uint2half_rd",
    "__uint2half_rn",
    "__uint2half_ru",
    "__uint2half_rz",
    "__ull2half_rd",
    "__ull2half_rn",
    "__ull2half_ru",
    "__ull2half_rz",
    "__ushort2half_rd",
    "__ushort2half_rn",
    "__ushort2half_ru",
    "__ushort2half_rz",
    "__half2char_rz",
    "__half2float",
    "__half2int_rd",
    "__half2int_rn",
    "__half2int_ru",
    "__half2int_rz",
    "__half2ll_rd",
    "__half2ll_rn",
    "__half2ll_ru",
    "__half2ll_rz",
    "__half2short_rd",
    "__half2short_rn",
    "__half2short_ru",
    "__half2short_rz",
    "__half2uchar_rz",
    "__half2uint_rd",
    "__half2uint_rn",
    "__half2uint_ru",
    "__half2uint_rz",
    "__half2ull_rd",
    "__half2ull_rn",
    "__half2ull_ru",
    "__half2ull_rz",
    "__half2ushort_rd",
    "__half2ushort_rn",
    "__half2ushort_ru",
    "__half2ushort_rz",
    "__short_as_half",
    "__ushort_as_half",
    "__half_as_short",
    "__half_as_ushort",
    "habs",
    "__habs",
    "hadd",
    "__hadd",
    "__hadd_rn",
    "__hadd_sat",
    "hdiv",
    "__hdiv",
    "heq",
    "__heq",
    "__hequ",
    "hfma",
    "__hfma",
    "__hfma_relu",
    "__hfma_sat",
    "hge",
    "__hge",
    "__hgeu",
    "hgt",
    "__hgt",
    "__hgtu",
    "__hisinf",
    "__hisnan",
    "hle",
    "__hle",
    "__hleu",
    "hlt",
    "__hlt",
    "__hltu",
    "hmax",
    "__hmax",
    "__hmax_nan",
    "hmin",
    "__hmin",
    "__hmin_nan",
    "hmul",
    "__hmul",
    "__hmul_rn",
    "__hmul_sat",
    "hne",
    "__hne",
    "hneg",
    "__hneg",
    "__hneu",
    "hsub",
    "__hsub",
    "__hsub_rn",
    "__hsub_sat",
    "atomicAdd",
    "hceil",
    "hcos",
    "hexp",
    "hexp10",
    "hexp2",
    "hfloor",
    "hlog",
    "hlog10",
    "hlog2",
    "hrcp",
    "hrint",
    "hrsqrt",
    "hsin",
    "hsqrt",
    "htanh",
    "htanh_approx",
    "htrunc",
]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.core.targetconfig import TargetConfig, Option

from numba.cuda.core.options import (
    ParallelOptions,
    FastMathOptions,
    InlineOptions,
)

from numba.cuda.core.callconv import BaseCallConv


class Flags(TargetConfig):
    __slots__ = ()

    enable_looplift = Option(
        type=bool,
        default=False,
        doc="Enable loop-lifting",
    )
    enable_pyobject = Option(
        type=bool,
        default=False,
        doc="Enable pyobject mode (in general)",
    )
    enable_pyobject_looplift = Option(
        type=bool,
        default=False,
        doc="Enable pyobject mode inside lifted loops",
    )
    enable_ssa = Option(
        type=bool,
        default=True,
        doc="Enable SSA",
    )
    force_pyobject = Option(
        type=bool,
        default=False,
        doc="Force pyobject mode inside the whole function",
    )
    release_gil = Option(
        type=bool,
        default=False,
        doc="Release GIL inside the native function",
    )
    no_compile = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    debuginfo = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    boundscheck = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    forceinline = Option(
        type=bool,
        default=False,
        doc="Force inlining of the function. Overrides _dbg_optnone.",
    )
    no_cpython_wrapper = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    no_cfunc_wrapper = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    auto_parallel = Option(
        type=ParallelOptions,
        default=ParallelOptions(False),
        doc="""Enable automatic parallel optimization, can be fine-tuned by
taking a dictionary of sub-options instead of a boolean, see parfor.py for
detail""",
    )
    nrt = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    no_rewrites = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    error_model = Option(
        type=str,
        default="python",
        doc="TODO",
    )
    fastmath = Option(
        type=FastMathOptions,
        default=FastMathOptions(False),
        doc="TODO",
    )
    noalias = Option(
        type=bool,
        default=False,
        doc="TODO",
    )
    inline = Option(
        type=InlineOptions,
        default=InlineOptions("never"),
        doc="TODO",
    )

    dbg_extend_lifetimes = Option(
        type=bool,
        default=False,
        doc=(
            "Extend variable lifetime for debugging. "
            "This automatically turns on with debug=True."
        ),
    )

    dbg_optnone = Option(
        type=bool,
        default=False,
        doc=(
            "Disable optimization for debug. "
            "Equivalent to adding optnone attribute in the LLVM Function."
        ),
    )

    dbg_directives_only = Option(
        type=bool,
        default=False,
        doc=(
            "Make debug emissions directives-only. "
            "Used when generating lineinfo."
        ),
    )


DEFAULT_FLAGS = Flags()
DEFAULT_FLAGS.nrt = True


def _nvvm_options_type(x):
    if x is None:
        return None

    else:
        assert isinstance(x, dict)
        return x


def _optional_int_type(x):
    if x is None:
        return None

    else:
        assert isinstance(x, int)
        return x


class CUDAFlags(Flags):
    nvvm_options = Option(
        type=_nvvm_options_type,
        default=None,
        doc="NVVM options",
    )
    compute_capability = Option(
        type=tuple,
        default=None,
        doc="Compute Capability",
    )
    max_registers = Option(
        type=_optional_int_type, default=None, doc="Max registers"
    )
    lto = Option(type=bool, default=False, doc="Enable Link-time Optimization")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from llvmlite import ir
from numba.cuda import types
from numba.cuda import cgutils
from numba.cuda.core.imputils import Registry
from numba.cuda import libdevice, libdevicefuncs

registry = Registry("libdeviceimpl")
lower = registry.lower


def libdevice_implement(func, retty, nbargs):
    def core(context, builder, sig, args):
        lmod = builder.module
        fretty = context.get_value_type(retty)
        fargtys = [context.get_value_type(arg.ty) for arg in nbargs]
        fnty = ir.FunctionType(fretty, fargtys)
        fn = cgutils.get_or_insert_function(lmod, fnty, func)
        return builder.call(fn, args)

    key = getattr(libdevice, func[5:])

    argtys = [arg.ty for arg in args if not arg.is_ptr]
    lower(key, *argtys)(core)


def libdevice_implement_multiple_returns(func, retty, prototype_args):
    sig = libdevicefuncs.create_signature(retty, prototype_args)
    nb_retty = sig.return_type

    def core(context, builder, sig, args):
        lmod = builder.module

        fargtys = []
        for arg in prototype_args:
            ty = context.get_value_type(arg.ty)
            if arg.is_ptr:
                ty = ty.as_pointer()
            fargtys.append(ty)

        fretty = context.get_value_type(retty)

        fnty = ir.FunctionType(fretty, fargtys)
        fn = cgutils.get_or_insert_function(lmod, fnty, func)

        # For returned values that are returned through a pointer, we need to
        # allocate variables on the stack and pass a pointer to them.
        actual_args = []
        virtual_args = []
        arg_idx = 0
        for arg in prototype_args:
            if arg.is_ptr:
                # Allocate space for return value and add to args
                tmp_arg = cgutils.alloca_once(
                    builder, context.get_value_type(arg.ty)
                )
                actual_args.append(tmp_arg)
                virtual_args.append(tmp_arg)
            else:
                actual_args.append(args[arg_idx])
                arg_idx += 1

        ret = builder.call(fn, actual_args)

        # Following the call, we need to assemble the returned values into a
        # tuple for returning back to the caller.
        tuple_args = []
        if retty != types.void:
            tuple_args.append(ret)
        tuple_args.extend(map(builder.load, virtual_args))

        if isinstance(nb_retty, types.UniTuple):
            return cgutils.pack_array(builder, tuple_args)
        else:
            return cgutils.pack_struct(builder, tuple_args)

    key = getattr(libdevice, func[5:])
    lower(key, *sig.args)(core)


for func, (retty, args) in libdevicefuncs.functions.items():
    if any([arg.is_ptr for arg in args]):
        libdevice_implement_multiple_returns(func, retty, args)
    else:
        libdevice_implement(func, retty, args)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
This is a direct translation of nvvm.h
"""

import logging
import re
import sys
import warnings
from ctypes import c_void_p, c_int, POINTER, c_char_p, c_size_t, byref, c_char

import threading

from llvmlite import ir

from .error import NvvmError, NvvmSupportError, NvvmWarning
from .libs import get_libdevice, open_libdevice, open_cudalib
from numba.cuda import cgutils


logger = logging.getLogger(__name__)

ADDRSPACE_GENERIC = 0
ADDRSPACE_GLOBAL = 1
ADDRSPACE_SHARED = 3
ADDRSPACE_CONSTANT = 4
ADDRSPACE_LOCAL = 5

# Opaque handle for compilation unit
nvvm_program = c_void_p

# Result code
nvvm_result = c_int

RESULT_CODE_NAMES = """
NVVM_SUCCESS
NVVM_ERROR_OUT_OF_MEMORY
NVVM_ERROR_PROGRAM_CREATION_FAILURE
NVVM_ERROR_IR_VERSION_MISMATCH
NVVM_ERROR_INVALID_INPUT
NVVM_ERROR_INVALID_PROGRAM
NVVM_ERROR_INVALID_IR
NVVM_ERROR_INVALID_OPTION
NVVM_ERROR_NO_MODULE_IN_PROGRAM
NVVM_ERROR_COMPILATION
""".split()

for i, k in enumerate(RESULT_CODE_NAMES):
    setattr(sys.modules[__name__], k, i)

_datalayout = (
    "e-p:64:64:64-i1:8:8-i8:8:8-i16:16:16-i32:32:32-i64:64:64-"
    "i128:128:128-f32:32:32-f64:64:64-v16:16:16-v32:32:32-"
    "v64:64:64-v128:128:128-n16:32:64"
)


def is_available():
    """
    Return if libNVVM is available
    """
    try:
        NVVM()
    except NvvmSupportError:
        return False
    else:
        return True


_nvvm_lock = threading.Lock()


class NVVM:
    """Process-wide singleton."""

    _PROTOTYPES = {
        # nvvmResult nvvmVersion(int *major, int *minor)
        "nvvmVersion": (nvvm_result, POINTER(c_int), POINTER(c_int)),
        # nvvmResult nvvmCreateProgram(nvvmProgram *cu)
        "nvvmCreateProgram": (nvvm_result, POINTER(nvvm_program)),
        # nvvmResult nvvmDestroyProgram(nvvmProgram *cu)
        "nvvmDestroyProgram": (nvvm_result, POINTER(nvvm_program)),
        # nvvmResult nvvmAddModuleToProgram(nvvmProgram cu, const char *buffer,
        #                                   size_t size, const char *name)
        "nvvmAddModuleToProgram": (
            nvvm_result,
            nvvm_program,
            c_char_p,
            c_size_t,
            c_char_p,
        ),
        # nvvmResult nvvmLazyAddModuleToProgram(nvvmProgram cu,
        #                                       const char* buffer,
        #                                       size_t size,
        #                                       const char *name)
        "nvvmLazyAddModuleToProgram": (
            nvvm_result,
            nvvm_program,
            c_char_p,
            c_size_t,
            c_char_p,
        ),
        # nvvmResult nvvmCompileProgram(nvvmProgram cu, int numOptions,
        #                          const char **options)
        "nvvmCompileProgram": (
            nvvm_result,
            nvvm_program,
            c_int,
            POINTER(c_char_p),
        ),
        # nvvmResult nvvmGetCompiledResultSize(nvvmProgram cu,
        #                                      size_t *bufferSizeRet)
        "nvvmGetCompiledResultSize": (
            nvvm_result,
            nvvm_program,
            POINTER(c_size_t),
        ),
        # nvvmResult nvvmGetCompiledResult(nvvmProgram cu, char *buffer)
        "nvvmGetCompiledResult": (nvvm_result, nvvm_program, c_char_p),
        # nvvmResult nvvmGetProgramLogSize(nvvmProgram cu,
        #                                      size_t *bufferSizeRet)
        "nvvmGetProgramLogSize": (nvvm_result, nvvm_program, POINTER(c_size_t)),
        # nvvmResult nvvmGetProgramLog(nvvmProgram cu, char *buffer)
        "nvvmGetProgramLog": (nvvm_result, nvvm_program, c_char_p),
        # nvvmResult nvvmIRVersion (int* majorIR, int* minorIR, int* majorDbg,
        #                           int* minorDbg )
        "nvvmIRVersion": (
            nvvm_result,
            POINTER(c_int),
            POINTER(c_int),
            POINTER(c_int),
            POINTER(c_int),
        ),
        # nvvmResult nvvmVerifyProgram (nvvmProgram prog, int numOptions,
        #                               const char** options)
        "nvvmVerifyProgram": (
            nvvm_result,
            nvvm_program,
            c_int,
            POINTER(c_char_p),
        ),
    }

    # Singleton reference
    __INSTANCE = None

    def __new__(cls):
        with _nvvm_lock:
            if cls.__INSTANCE is None:
                cls.__INSTANCE = inst = object.__new__(cls)
                try:
                    inst.driver = open_cudalib("nvvm")
                except OSError as e:
                    cls.__INSTANCE = None
                    errmsg = "libNVVM cannot be found. Please install the cuda-toolkit conda package:\n%s"
                    raise NvvmSupportError(errmsg % e)

                # Find & populate functions
                for name, proto in inst._PROTOTYPES.items():
                    func = getattr(inst.driver, name)
                    func.restype = proto[0]
                    func.argtypes = proto[1:]
                    setattr(inst, name, func)

        return cls.__INSTANCE

    def __init__(self):
        ir_versions = self.get_ir_version()
        self._majorIR = ir_versions[0]
        self._minorIR = ir_versions[1]
        self._majorDbg = ir_versions[2]
        self._minorDbg = ir_versions[3]

    @property
    def data_layout(self):
        return _datalayout

    def get_version(self):
        major = c_int()
        minor = c_int()
        err = self.nvvmVersion(byref(major), byref(minor))
        self.check_error(err, "Failed to get version.")
        return major.value, minor.value

    def get_ir_version(self):
        majorIR = c_int()
        minorIR = c_int()
        majorDbg = c_int()
        minorDbg = c_int()
        err = self.nvvmIRVersion(
            byref(majorIR), byref(minorIR), byref(majorDbg), byref(minorDbg)
        )
        self.check_error(err, "Failed to get IR version.")
        return majorIR.value, minorIR.value, majorDbg.value, minorDbg.value

    def check_options(self, options):
        """
        Check if the specified options are supported by the current libNVVM version.

        The options are a list of bytes, each representing a compiler option.

        If the test program fails to compile, the options are not supported and False
        is returned.

        If the test program compiles successfully, True is returned.
        """

        precheck_nvvm_ir = """target triple = "nvptx64-unknown-cuda"
        target datalayout = "e-p:64:64:64-i1:8:8-i8:8:8-i16:16:16-i32:32:32-i64:64:64-i128:128:128-f32:32:32-f64:64:64-v16:16:16-v32:32:32-v64:64:64-v128:128:128-n16:32:64"

        define void @dummy_kernel() {{
        entry:
        ret void
        }}

        !nvvm.annotations = !{{!0}}
        !0 = !{{void ()* @dummy_kernel, !"kernel", i32 1}}

        !nvvmir.version = !{{!1}}
        !1 = !{{i32 {major}, i32 {minor}, i32 {debug_major}, i32 {debug_minor}}}
        """

        # Use a test program to compile in order to determine
        # if the provided options are supported by libnvvm.
        program = c_void_p()
        try:
            # Create the NVVM program
            err = self.nvvmCreateProgram(byref(program))
            self.check_error(err, "Failed to create test program.")

            # Add the test program to the compilation unit
            precheck_nvvm_ir = precheck_nvvm_ir.format(
                major=self._majorIR,
                minor=self._minorIR,
                debug_major=self._majorDbg,
                debug_minor=self._minorDbg,
            )
            precheck_ir_bytes = precheck_nvvm_ir.encode("utf-8")
            err = self.nvvmAddModuleToProgram(
                program,
                precheck_ir_bytes,
                len(precheck_ir_bytes),
                "precheck.ll".encode("utf-8"),
            )
            self.check_error(err, "Failed to add test module.")

            # Compile the test program
            option_ptrs = (c_char_p * len(options))(
                *[c_char_p(x) for x in options]
            )
            err = self.nvvmVerifyProgram(program, len(options), option_ptrs)
            if err != 0:
                return False
            err = self.nvvmCompileProgram(program, len(options), option_ptrs)
            if err != 0:
                return False
        except Exception:
            return False
        finally:
            if program.value:
                # Destroy the NVVM program, not fatal if it fails
                self.nvvmDestroyProgram(byref(program))
        return True

    def check_error(self, error, msg, exit=False):
        if error:
            exc = NvvmError(msg, RESULT_CODE_NAMES[error])
            if exit:
                print(exc)
                sys.exit(1)
            else:
                raise exc


class CompilationUnit:
    """
    A CompilationUnit is a set of LLVM modules that are compiled to PTX or
    LTO-IR with NVVM.

    Compilation options are accepted as a dict mapping option names to values,
    with the following considerations:

    - Underscores (`_`) in option names are converted to dashes (`-`), to match
      NVVM's option name format.
    - Options that take a value will be emitted in the form "-<name>=<value>".
    - Booleans passed as option values will be converted to integers.
    - Options which take no value (such as `-gen-lto`) should have a value of
      `None` and will be emitted in the form "-<name>".

    For documentation on NVVM compilation options, see the CUDA Toolkit
    Documentation:

    https://docs.nvidia.com/cuda/libnvvm-api/index.html#_CPPv418nvvmCompileProgram11nvvmProgramiPPKc
    """

    def __init__(self, options):
        self.driver = NVVM()
        self._handle = nvvm_program()
        err = self.driver.nvvmCreateProgram(byref(self._handle))
        self.driver.check_error(err, "Failed to create CU")

        def stringify_option(k, v):
            k = k.replace("_", "-")

            if v is None:
                return f"-{k}".encode("utf-8")

            if isinstance(v, bool):
                v = int(v)

            return f"-{k}={v}".encode("utf-8")

        debug_build = "g" in options
        options = [stringify_option(k, v) for k, v in options.items()]

        # If the -numba-debug option is supported by libnvvm, pass it to the
        # compiler for debug builds.
        if debug_build:
            numba_debug_option = stringify_option("numba-debug", None)
            if self.driver.check_options(options + [numba_debug_option]):
                options.append(numba_debug_option)

        option_ptrs = (c_char_p * len(options))(*[c_char_p(x) for x in options])

        # We keep both the options and the pointers to them so that options are
        # not destroyed before we've used their values
        self.options = options
        self.option_ptrs = option_ptrs
        self.n_options = len(options)

    def __del__(self):
        driver = NVVM()
        err = driver.nvvmDestroyProgram(byref(self._handle))
        driver.check_error(err, "Failed to destroy CU", exit=True)

    def add_module(self, buffer):
        """
        Add a module level NVVM IR to a compilation unit.
        - The buffer should contain an NVVM module IR either in the bitcode
          representation (LLVM3.0) or in the text representation.
        """
        err = self.driver.nvvmAddModuleToProgram(
            self._handle, buffer, len(buffer), None
        )
        self.driver.check_error(err, "Failed to add module")

    def lazy_add_module(self, buffer):
        """
        Lazily add an NVVM IR module to a compilation unit.
        The buffer should contain NVVM module IR either in the bitcode
        representation or in the text representation.
        """
        err = self.driver.nvvmLazyAddModuleToProgram(
            self._handle, buffer, len(buffer), None
        )
        self.driver.check_error(err, "Failed to add module")

    def verify(self):
        """
        Run the NVVM verifier on all code added to the compilation unit.
        """
        err = self.driver.nvvmVerifyProgram(
            self._handle, self.n_options, self.option_ptrs
        )
        self._try_error(err, "Failed to verify\n")

    def compile(self):
        """
        Compile all modules added to the compilation unit and return the
        resulting PTX or LTO-IR (depending on the options).
        """
        err = self.driver.nvvmCompileProgram(
            self._handle, self.n_options, self.option_ptrs
        )
        self._try_error(err, "Failed to compile\n")

        # Get result
        result_size = c_size_t()
        err = self.driver.nvvmGetCompiledResultSize(
            self._handle, byref(result_size)
        )

        self._try_error(err, "Failed to get size of compiled result.")

        output_buffer = (c_char * result_size.value)()
        err = self.driver.nvvmGetCompiledResult(self._handle, output_buffer)
        self._try_error(err, "Failed to get compiled result.")

        # Get log
        self.log = self.get_log()
        if self.log:
            warnings.warn(self.log, category=NvvmWarning)

        return output_buffer[:]

    def _try_error(self, err, msg):
        self.driver.check_error(err, "%s\n%s" % (msg, self.get_log()))

    def get_log(self):
        reslen = c_size_t()
        err = self.driver.nvvmGetProgramLogSize(self._handle, byref(reslen))
        self.driver.check_error(err, "Failed to get compilation log size.")

        if reslen.value > 1:
            logbuf = (c_char * reslen.value)()
            err = self.driver.nvvmGetProgramLog(self._handle, logbuf)
            self.driver.check_error(err, "Failed to get compilation log.")

            return logbuf.value.decode("utf8")  # populate log attribute

        return ""


MISSING_LIBDEVICE_FILE_MSG = """Missing libdevice file.
``cuda-nvcc`` and ``cuda-nvrtc`` are required:

    $ conda install -c conda-forge cuda-nvcc cuda-nvrtc "cuda-version>=12.0"
"""


class LibDevice:
    _cache_ = None

    def __init__(self):
        if self._cache_ is None:
            if get_libdevice() is None:
                raise RuntimeError(MISSING_LIBDEVICE_FILE_MSG)
            self._cache_ = open_libdevice()

        self.bc = self._cache_

    def get(self):
        return self.bc


cas_nvvm = """
    %cas_success = cmpxchg volatile {Ti}* %iptr, {Ti} %old, {Ti} %new monotonic monotonic
    %cas = extractvalue {{ {Ti}, i1 }} %cas_success, 0
"""  # noqa: E501


# Translation of code from CUDA Programming Guide v6.5, section B.12
ir_numba_atomic_binary_template = """
define internal {T} @___numba_atomic_{T}_{FUNC}({T}* %ptr, {T} %val) alwaysinline {{
entry:
    %iptr = bitcast {T}* %ptr to {Ti}*
    %old2 = load volatile {Ti}, {Ti}* %iptr
    br label %attempt

attempt:
    %old = phi {Ti} [ %old2, %entry ], [ %cas, %attempt ]
    %dold = bitcast {Ti} %old to {T}
    %dnew = {OP} {T} %dold, %val
    %new = bitcast {T} %dnew to {Ti}
    {CAS}
    %repeat = icmp ne {Ti} %cas, %old
    br i1 %repeat, label %attempt, label %done

done:
    %result = bitcast {Ti} %old to {T}
    ret {T} %result
}}
"""  # noqa: E501

ir_numba_atomic_inc_template = """
define internal {T} @___numba_atomic_{Tu}_inc({T}* %iptr, {T} %val) alwaysinline {{
entry:
    %old2 = load volatile {T}, {T}* %iptr
    br label %attempt

attempt:
    %old = phi {T} [ %old2, %entry ], [ %cas, %attempt ]
    %bndchk = icmp ult {T} %old, %val
    %inc = add {T} %old, 1
    %new = select i1 %bndchk, {T} %inc, {T} 0
    {CAS}
    %repeat = icmp ne {T} %cas, %old
    br i1 %repeat, label %attempt, label %done

done:
    ret {T} %old
}}
"""  # noqa: E501

ir_numba_atomic_dec_template = """
define internal {T} @___numba_atomic_{Tu}_dec({T}* %iptr, {T} %val) alwaysinline {{
entry:
    %old2 = load volatile {T}, {T}* %iptr
    br label %attempt

attempt:
    %old = phi {T} [ %old2, %entry ], [ %cas, %attempt ]
    %dec = add {T} %old, -1
    %bndchk = icmp ult {T} %dec, %val
    %new = select i1 %bndchk, {T} %dec, {T} %val
    {CAS}
    %repeat = icmp ne {T} %cas, %old
    br i1 %repeat, label %attempt, label %done

done:
    ret {T} %old
}}
"""  # noqa: E501

ir_numba_atomic_minmax_template = """
define internal {T} @___numba_atomic_{T}_{NAN}{FUNC}({T}* %ptr, {T} %val) alwaysinline {{
entry:
    %ptrval = load volatile {T}, {T}* %ptr
    ; Return early when:
    ; - For nanmin / nanmax when val is a NaN
    ; - For min / max when val or ptr is a NaN
    %early_return = fcmp uno {T} %val, %{PTR_OR_VAL}val
    br i1 %early_return, label %done, label %lt_check

lt_check:
    %dold = phi {T} [ %ptrval, %entry ], [ %dcas, %attempt ]
    ; Continue attempts if dold less or greater than val (depending on whether min or max)
    ; or if dold is NaN (for nanmin / nanmax)
    %cmp = fcmp {OP} {T} %dold, %val
    br i1 %cmp, label %attempt, label %done

attempt:
    ; Attempt to swap in the value
    %old = bitcast {T} %dold to {Ti}
    %iptr = bitcast {T}* %ptr to {Ti}*
    %new = bitcast {T} %val to {Ti}
    {CAS}
    %dcas = bitcast {Ti} %cas to {T}
    br label %lt_check

done:
    ret {T} %ptrval
}}
"""  # noqa: E501


def ir_cas(Ti):
    return cas_nvvm.format(Ti=Ti)


def ir_numba_atomic_binary(T, Ti, OP, FUNC):
    params = dict(T=T, Ti=Ti, OP=OP, FUNC=FUNC, CAS=ir_cas(Ti))
    return ir_numba_atomic_binary_template.format(**params)


def ir_numba_atomic_minmax(T, Ti, NAN, OP, PTR_OR_VAL, FUNC):
    params = dict(
        T=T,
        Ti=Ti,
        NAN=NAN,
        OP=OP,
        PTR_OR_VAL=PTR_OR_VAL,
        FUNC=FUNC,
        CAS=ir_cas(Ti),
    )

    return ir_numba_atomic_minmax_template.format(**params)


def ir_numba_atomic_inc(T, Tu):
    return ir_numba_atomic_inc_template.format(T=T, Tu=Tu, CAS=ir_cas(T))


def ir_numba_atomic_dec(T, Tu):
    return ir_numba_atomic_dec_template.format(T=T, Tu=Tu, CAS=ir_cas(T))


def llvm_replace(llvmir):
    replacements = [
        (
            'declare double @"___numba_atomic_double_add"(double* %".1", double %".2")',  # noqa: E501
            ir_numba_atomic_binary(T="double", Ti="i64", OP="fadd", FUNC="add"),
        ),
        (
            'declare float @"___numba_atomic_float_sub"(float* %".1", float %".2")',  # noqa: E501
            ir_numba_atomic_binary(T="float", Ti="i32", OP="fsub", FUNC="sub"),
        ),
        (
            'declare double @"___numba_atomic_double_sub"(double* %".1", double %".2")',  # noqa: E501
            ir_numba_atomic_binary(T="double", Ti="i64", OP="fsub", FUNC="sub"),
        ),
        (
            'declare i64 @"___numba_atomic_u64_inc"(i64* %".1", i64 %".2")',
            ir_numba_atomic_inc(T="i64", Tu="u64"),
        ),
        (
            'declare i64 @"___numba_atomic_u64_dec"(i64* %".1", i64 %".2")',
            ir_numba_atomic_dec(T="i64", Tu="u64"),
        ),
        (
            'declare float @"___numba_atomic_float_max"(float* %".1", float %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="float",
                Ti="i32",
                NAN="",
                OP="nnan olt",
                PTR_OR_VAL="ptr",
                FUNC="max",
            ),
        ),
        (
            'declare double @"___numba_atomic_double_max"(double* %".1", double %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="double",
                Ti="i64",
                NAN="",
                OP="nnan olt",
                PTR_OR_VAL="ptr",
                FUNC="max",
            ),
        ),
        (
            'declare float @"___numba_atomic_float_min"(float* %".1", float %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="float",
                Ti="i32",
                NAN="",
                OP="nnan ogt",
                PTR_OR_VAL="ptr",
                FUNC="min",
            ),
        ),
        (
            'declare double @"___numba_atomic_double_min"(double* %".1", double %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="double",
                Ti="i64",
                NAN="",
                OP="nnan ogt",
                PTR_OR_VAL="ptr",
                FUNC="min",
            ),
        ),
        (
            'declare float @"___numba_atomic_float_nanmax"(float* %".1", float %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="float",
                Ti="i32",
                NAN="nan",
                OP="ult",
                PTR_OR_VAL="",
                FUNC="max",
            ),
        ),
        (
            'declare double @"___numba_atomic_double_nanmax"(double* %".1", double %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="double",
                Ti="i64",
                NAN="nan",
                OP="ult",
                PTR_OR_VAL="",
                FUNC="max",
            ),
        ),
        (
            'declare float @"___numba_atomic_float_nanmin"(float* %".1", float %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="float",
                Ti="i32",
                NAN="nan",
                OP="ugt",
                PTR_OR_VAL="",
                FUNC="min",
            ),
        ),
        (
            'declare double @"___numba_atomic_double_nanmin"(double* %".1", double %".2")',  # noqa: E501
            ir_numba_atomic_minmax(
                T="double",
                Ti="i64",
                NAN="nan",
                OP="ugt",
                PTR_OR_VAL="",
                FUNC="min",
            ),
        ),
        ("immarg", ""),
    ]

    for decl, fn in replacements:
        llvmir = llvmir.replace(decl, fn)

    llvmir = llvm150_to_70_ir(llvmir)

    return llvmir


def compile_ir(llvmir, **options):
    if isinstance(llvmir, str):
        llvmir = [llvmir]

    if options.pop("fastmath", False):
        options.update(
            {
                "ftz": True,
                "fma": True,
                "prec_div": False,
                "prec_sqrt": False,
            }
        )

    cu = CompilationUnit(options)

    for mod in llvmir:
        mod = llvm_replace(mod)
        cu.add_module(mod.encode("utf8"))
    cu.verify()

    # We add libdevice following verification so that it is not subject to the
    # verifier's requirements
    libdevice = LibDevice()
    cu.lazy_add_module(libdevice.get())

    return cu.compile()


re_attributes_def = re.compile(r"^attributes #\d+ = \{ ([\w\s]+)\ }")


def llvm150_to_70_ir(ir):
    """
    Convert LLVM 15.0 IR for LLVM 7.0.
    """
    buf = []
    for line in ir.splitlines():
        if line.startswith("attributes #"):
            # Remove function attributes unsupported by LLVM 7.0
            m = re_attributes_def.match(line)
            attrs = m.group(1).split()
            attrs = " ".join(a for a in attrs if a != "willreturn")
            line = line.replace(m.group(1), attrs)

        buf.append(line)

    return "\n".join(buf)


def set_cuda_kernel(function):
    """
    Mark a function as a CUDA kernel. Kernels have the following requirements:

    - Metadata that marks them as a kernel.
    - Addition to the @llvm.used list, so that they will not be discarded.
    - The noinline attribute is not permitted, because this causes NVVM to emit
      a warning, which counts as failing IR verification.

    Presently it is assumed that there is one kernel per module, which holds
    for Numba-jitted functions. If this changes in future or this function is
    to be used externally, this function may need modification to add to the
    @llvm.used list rather than creating it.
    """
    module = function.module

    # Add kernel metadata
    mdstr = ir.MetaDataString(module, "kernel")
    mdvalue = ir.Constant(ir.IntType(32), 1)
    md = module.add_metadata((function, mdstr, mdvalue))

    nmd = cgutils.get_or_insert_named_metadata(module, "nvvm.annotations")
    nmd.add(md)

    # Create the used list
    ptrty = ir.IntType(8).as_pointer()
    usedty = ir.ArrayType(ptrty, 1)

    fnptr = function.bitcast(ptrty)

    llvm_used = ir.GlobalVariable(module, usedty, "llvm.used")
    llvm_used.linkage = "appending"
    llvm_used.section = "llvm.metadata"
    llvm_used.initializer = ir.Constant(usedty, [fnptr])

    # Remove 'noinline' if it is present.
    function.attributes.discard("noinline")


def set_launch_bounds(kernel, launch_bounds):
    # Based on: CUDA C / C++ Programming Guide 12.9, Section 8.38:
    # https://docs.nvidia.com/cuda/archive/12.9.0/cuda-c-programming-guide/index.html#launch-bounds
    # PTX ISA Specification Version 8.7, Section 11.4:
    # https://docs.nvidia.com/cuda/archive/12.8.1/parallel-thread-execution/index.html#performance-tuning-directives
    # NVVM IR Specification 12.9, Section 13:
    # https://docs.nvidia.com/cuda/archive/12.9.0/nvvm-ir-spec/index.html#global-property-annotation

    if launch_bounds is None:
        return

    if isinstance(launch_bounds, int):
        launch_bounds = (launch_bounds,)

    if (n := len(launch_bounds)) > 3:
        raise ValueError(
            f"Got {n} launch bounds: {launch_bounds}. A maximum of three are supported: "
            "(max_threads_per_block, min_blocks_per_sm, max_blocks_per_cluster)"
        )

    module = kernel.module
    nvvm_annotations = cgutils.get_or_insert_named_metadata(
        module, "nvvm.annotations"
    )

    # Note that only maxntidx is used even though NVVM IR and PTX allow
    # maxntidy and maxntidz. This is because the thread block size limit
    # pertains only to the total number of threads, and therefore bounds on
    # individual dimensions may be exceeded anyway. To prevent an unsurprising
    # interface, it is cleaner to only allow setting total size via maxntidx
    # and assuming y and z to be 1 (as is the case in CUDA C/C++).

    properties = (
        # Max threads per block
        "maxntidx",
        # Min blocks per multiprocessor
        "minctasm",
        # Max blocks per cluster
        "cluster_max_blocks",
    )

    for prop, bound in zip(properties, launch_bounds):
        mdstr = ir.MetaDataString(module, prop)
        mdvalue = ir.Constant(ir.IntType(32), bound)
        md = module.add_metadata((kernel, mdstr, mdvalue))
        nvvm_annotations.add(md)


def add_ir_version(mod):
    """Add NVVM IR version to module"""
    # We specify the IR version to match the current NVVM's IR version
    i32 = ir.IntType(32)
    ir_versions = [i32(v) for v in NVVM().get_ir_version()]
    md_ver = mod.add_metadata(ir_versions)
    mod.add_named_metadata("nvvmir.version", md_ver)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Utilities to simplify the boilerplate for native lowering.
"""

import collections
import contextlib
from enum import Enum

from numba.cuda import typing, cgutils
from numba.cuda import types
from numba.cuda.core import callconv
from numba.cuda.typing.templates import BaseRegistryLoader


class Registry:
    """
    A registry of function and attribute implementations.
    """

    def __init__(self, name="unspecified"):
        self.name = name
        self.functions = []
        self.getattrs = []
        self.setattrs = []
        self.casts = []
        self.constants = []

    def lower(self, func, *argtys):
        """
        Decorate an implementation of *func* for the given argument types.
        *func* may be an actual global function object, or any
        pseudo-function supported by Numba, such as "getitem".

        The decorated implementation has the signature
        (context, builder, sig, args).
        """

        def decorate(impl):
            self.functions.append((impl, func, argtys))
            return impl

        return decorate

    def _decorate_attr(self, impl, ty, attr, impl_list, decorator):
        real_impl = decorator(impl, ty, attr)
        impl_list.append((real_impl, attr, real_impl.signature))
        return impl

    def lower_getattr(self, ty, attr):
        """
        Decorate an implementation of __getattr__ for type *ty* and
        the attribute *attr*.

        The decorated implementation will have the signature
        (context, builder, typ, val).
        """

        def decorate(impl):
            return self._decorate_attr(
                impl, ty, attr, self.getattrs, _decorate_getattr
            )

        return decorate

    def lower_getattr_generic(self, ty):
        """
        Decorate the fallback implementation of __getattr__ for type *ty*.

        The decorated implementation will have the signature
        (context, builder, typ, val, attr).  The implementation is
        called for attributes which haven't been explicitly registered
        with lower_getattr().
        """
        return self.lower_getattr(ty, None)

    def lower_setattr(self, ty, attr):
        """
        Decorate an implementation of __setattr__ for type *ty* and
        the attribute *attr*.

        The decorated implementation will have the signature
        (context, builder, sig, args).
        """

        def decorate(impl):
            return self._decorate_attr(
                impl, ty, attr, self.setattrs, _decorate_setattr
            )

        return decorate

    def lower_setattr_generic(self, ty):
        """
        Decorate the fallback implementation of __setattr__ for type *ty*.

        The decorated implementation will have the signature
        (context, builder, sig, args, attr).  The implementation is
        called for attributes which haven't been explicitly registered
        with lower_setattr().
        """
        return self.lower_setattr(ty, None)

    def lower_cast(self, fromty, toty):
        """
        Decorate the implementation of implicit conversion between
        *fromty* and *toty*.

        The decorated implementation will have the signature
        (context, builder, fromty, toty, val).
        """

        def decorate(impl):
            self.casts.append((impl, (fromty, toty)))
            return impl

        return decorate

    def lower_constant(self, ty):
        """
        Decorate the implementation for creating a constant of type *ty*.

        The decorated implementation will have the signature
        (context, builder, ty, pyval).
        """

        def decorate(impl):
            self.constants.append((impl, (ty,)))
            return impl

        return decorate

    def __repr__(self):
        return f"Lowering Registry<{self.name}>"


class RegistryLoader(BaseRegistryLoader):
    """
    An incremental loader for a target registry.
    """

    registry_items = ("functions", "getattrs", "setattrs", "casts", "constants")


# Global registry for implementations of builtin operations
# (functions, attributes, type casts)
builtin_registry = Registry("builtin_registry")

lower_builtin = builtin_registry.lower
lower_getattr = builtin_registry.lower_getattr
lower_getattr_generic = builtin_registry.lower_getattr_generic
lower_setattr = builtin_registry.lower_setattr
lower_setattr_generic = builtin_registry.lower_setattr_generic
lower_cast = builtin_registry.lower_cast
lower_constant = builtin_registry.lower_constant


def _decorate_getattr(impl, ty, attr):
    real_impl = impl

    if attr is not None:

        def res(context, builder, typ, value, attr):
            return real_impl(context, builder, typ, value)
    else:

        def res(context, builder, typ, value, attr):
            return real_impl(context, builder, typ, value, attr)

    res.signature = (ty,)
    res.attr = attr
    return res


def _decorate_setattr(impl, ty, attr):
    real_impl = impl

    if attr is not None:

        def res(context, builder, sig, args, attr):
            return real_impl(context, builder, sig, args)
    else:

        def res(context, builder, sig, args, attr):
            return real_impl(context, builder, sig, args, attr)

    res.signature = (ty, types.Any)
    res.attr = attr
    return res


def fix_returning_optional(context, builder, sig, status, retval):
    # Reconstruct optional return type
    if isinstance(sig.return_type, types.Optional):
        value_type = sig.return_type.type
        optional_none = context.make_optional_none(builder, value_type)
        retvalptr = cgutils.alloca_once_value(builder, optional_none)
        with builder.if_then(builder.not_(status.is_none)):
            optional_value = context.make_optional_value(
                builder,
                value_type,
                retval,
            )
            builder.store(optional_value, retvalptr)
        retval = builder.load(retvalptr)
    return retval


def user_function(fndesc, libs):
    """
    A wrapper inserting code calling Numba-compiled *fndesc*.
    """

    def imp(context, builder, sig, args, fndesc=fndesc):
        func = fndesc.declare_function(builder.module)
        # env=None assumes this is a nopython function
        status, retval = fndesc.call_conv.call_function(
            builder, func, fndesc.restype, fndesc.argtypes, args
        )

        if status is not None:
            with cgutils.if_unlikely(builder, status.is_error):
                fndesc.call_conv.return_status_propagate(builder, status)

        assert sig.return_type == fndesc.restype

        # Reconstruct optional return type
        retval = fix_returning_optional(context, builder, sig, status, retval)
        # If the data representations don't match up
        if retval.type != context.get_value_type(sig.return_type):
            msg = "function returned {0} but expect {1}"
            raise TypeError(msg.format(retval.type, sig.return_type))

        return impl_ret_new_ref(context, builder, fndesc.restype, retval)

    imp.signature = fndesc.argtypes
    imp.libs = tuple(libs)
    return imp


def user_generator(gendesc, libs):
    """
    A wrapper inserting code calling Numba-compiled *gendesc*.
    """

    def imp(context, builder, sig, args, gendesc=gendesc):
        func = gendesc.declare_function(builder.module)
        # env=None assumes this is a nopython function
        status, retval = gendesc.call_conv.call_function(
            builder, func, gendesc.restype, gendesc.argtypes, args
        )
        # Return raw status for caller to process StopIteration
        return status, retval

    imp.libs = tuple(libs)
    return imp


def iterator_impl(iterable_type, iterator_type):
    """
    Decorator a given class as implementing *iterator_type*
    (by providing an `iternext()` method).
    """

    def wrapper(cls):
        # These are unbound methods
        iternext = cls.iternext

        @iternext_impl(RefType.BORROWED)
        def iternext_wrapper(context, builder, sig, args, result):
            (value,) = args
            iterobj = cls(context, builder, value)
            return iternext(iterobj, context, builder, result)

        lower_builtin("iternext", iterator_type)(iternext_wrapper)
        return cls

    return wrapper


class _IternextResult:
    """
    A result wrapper for iteration, passed by iternext_impl() into the
    wrapped function.
    """

    __slots__ = ("_context", "_builder", "_pairobj")

    def __init__(self, context, builder, pairobj):
        self._context = context
        self._builder = builder
        self._pairobj = pairobj

    def set_exhausted(self):
        """
        Mark the iterator as exhausted.
        """
        self._pairobj.second = self._context.get_constant(types.boolean, False)

    def set_valid(self, is_valid=True):
        """
        Mark the iterator as valid according to *is_valid* (which must
        be either a Python boolean or a LLVM inst).
        """
        if is_valid in (False, True):
            is_valid = self._context.get_constant(types.boolean, is_valid)
        self._pairobj.second = is_valid

    def yield_(self, value):
        """
        Mark the iterator as yielding the given *value* (a LLVM inst).
        """
        self._pairobj.first = value

    def is_valid(self):
        """
        Return whether the iterator is marked valid.
        """
        return self._context.get_argument_value(
            self._builder, types.boolean, self._pairobj.second
        )

    def yielded_value(self):
        """
        Return the iterator's yielded value, if any.
        """
        return self._pairobj.first


class RefType(Enum):
    """
    Enumerate the reference type
    """

    """
    A new reference
    """
    NEW = 1
    """
    A borrowed reference
    """
    BORROWED = 2
    """
    An untracked reference
    """
    UNTRACKED = 3


def iternext_impl(ref_type=None):
    """
    Wrap the given iternext() implementation so that it gets passed
    an _IternextResult() object easing the returning of the iternext()
    result pair.

    ref_type: a numba.targets.imputils.RefType value, the reference type used is
    that specified through the RefType enum.

    The wrapped function will be called with the following signature:
        (context, builder, sig, args, iternext_result)
    """
    if ref_type not in [x for x in RefType]:
        raise ValueError("ref_type must be an enum member of imputils.RefType")

    def outer(func):
        def wrapper(context, builder, sig, args):
            pair_type = sig.return_type
            pairobj = context.make_helper(builder, pair_type)
            func(
                context,
                builder,
                sig,
                args,
                _IternextResult(context, builder, pairobj),
            )
            if ref_type == RefType.NEW:
                impl_ret = impl_ret_new_ref
            elif ref_type == RefType.BORROWED:
                impl_ret = impl_ret_borrowed
            elif ref_type == RefType.UNTRACKED:
                impl_ret = impl_ret_untracked
            else:
                raise ValueError("Unknown ref_type encountered")
            return impl_ret(context, builder, pair_type, pairobj._getvalue())

        return wrapper

    return outer


def call_getiter(context, builder, iterable_type, val):
    """
    Call the `getiter()` implementation for the given *iterable_type*
    of value *val*, and return the corresponding LLVM inst.
    """
    getiter_sig = typing.signature(iterable_type.iterator_type, iterable_type)
    getiter_impl = context.get_function("getiter", getiter_sig)
    return getiter_impl(builder, (val,))


def call_iternext(context, builder, iterator_type, val):
    """
    Call the `iternext()` implementation for the given *iterator_type*
    of value *val*, and return a convenience _IternextResult() object
    reflecting the results.
    """
    itemty = iterator_type.yield_type
    pair_type = types.Pair(itemty, types.boolean)
    iternext_sig = typing.signature(pair_type, iterator_type)
    iternext_impl = context.get_function("iternext", iternext_sig)
    val = iternext_impl(builder, (val,))
    pairobj = context.make_helper(builder, pair_type, val)
    return _IternextResult(context, builder, pairobj)


def call_len(context, builder, ty, val):
    """
    Call len() on the given value.  Return None if len() isn't defined on
    this type.
    """
    try:
        len_impl = context.get_function(
            len,
            typing.signature(
                types.intp,
                ty,
            ),
        )
    except NotImplementedError:
        return None
    else:
        return len_impl(builder, (val,))


_ForIterLoop = collections.namedtuple("_ForIterLoop", ("value", "do_break"))


@contextlib.contextmanager
def for_iter(context, builder, iterable_type, val):
    """
    Simulate a for loop on the given iterable.  Yields a namedtuple with
    the given members:
    - `value` is the value being yielded
    - `do_break` is a callable to early out of the loop
    """
    iterator_type = iterable_type.iterator_type
    iterval = call_getiter(context, builder, iterable_type, val)

    bb_body = builder.append_basic_block("for_iter.body")
    bb_end = builder.append_basic_block("for_iter.end")

    def do_break():
        builder.branch(bb_end)

    builder.branch(bb_body)

    with builder.goto_block(bb_body):
        res = call_iternext(context, builder, iterator_type, iterval)
        with builder.if_then(builder.not_(res.is_valid()), likely=False):
            builder.branch(bb_end)
        yield _ForIterLoop(res.yielded_value(), do_break)
        builder.branch(bb_body)

    builder.position_at_end(bb_end)
    if context.enable_nrt:
        context.nrt.decref(builder, iterator_type, iterval)


def impl_ret_new_ref(ctx, builder, retty, ret):
    """
    The implementation returns a new reference.
    """
    return ret


def impl_ret_borrowed(ctx, builder, retty, ret):
    """
    The implementation returns a borrowed reference.
    This function automatically incref so that the implementation is
    returning a new reference.
    """
    if ctx.enable_nrt:
        ctx.nrt.incref(builder, retty, ret)
    return ret


def impl_ret_untracked(ctx, builder, retty, ret):
    """
    The return type is not a NRT object.
    """
    return ret


@contextlib.contextmanager
def force_error_model(context, model_name="numpy"):
    """
    Temporarily change the context's error model.
    """
    old_error_model = context.error_model
    context.error_model = callconv.create_error_model(model_name, context)
    try:
        yield
    finally:
        context.error_model = old_error_model


def numba_typeref_ctor(*args, **kwargs):
    """A stub for use internally by Numba when a call is emitted
    on a TypeRef.
    """
    raise NotImplementedError("This function should not be executed.")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.core import errors
from numba.cuda import types
from numba.cuda.typing.npydecl import (
    parse_dtype,
    parse_shape,
)
from numba.cuda.typing.templates import (
    AttributeTemplate,
    ConcreteTemplate,
    AbstractTemplate,
    CallableTemplate,
    signature,
    Registry,
)
from numba.cuda.types.ext_types import dim3
from numba import cuda

registry = Registry()
register = registry.register
register_attr = registry.register_attr
register_global = registry.register_global


class Cuda_array_decl(CallableTemplate):
    def generic(self):
        def typer(shape, dtype, alignment=None):
            # Only integer literals and tuples of integer literals are valid
            # shapes
            if isinstance(shape, types.Integer):
                if not isinstance(shape, types.IntegerLiteral):
                    return None
            elif isinstance(shape, (types.Tuple, types.UniTuple)):
                if any(
                    [not isinstance(s, types.IntegerLiteral) for s in shape]
                ):
                    return None
            else:
                return None

            if alignment is not None:
                permitted = (types.IntegerLiteral, types.NoneType)
                if not isinstance(alignment, permitted):
                    msg = "alignment must be a constant integer"
                    raise errors.RequireLiteralValue(msg)

            # N.B. We don't use alignment for typing; it's not part of
            #      types.Array.  The value supplied to the array declaration
            #      is handled in the lowering.

            ndim = parse_shape(shape)
            nb_dtype = parse_dtype(dtype)
            if nb_dtype is not None and ndim is not None:
                return types.Array(dtype=nb_dtype, ndim=ndim, layout="C")

        return typer


@register
class Cuda_shared_array(Cuda_array_decl):
    key = cuda.shared.array


@register
class Cuda_local_array(Cuda_array_decl):
    key = cuda.local.array


@register
class Cuda_const_array_like(CallableTemplate):
    key = cuda.const.array_like

    def generic(self):
        def typer(ndarray):
            return ndarray

        return typer


@register
class Cuda_threadfence_device(ConcreteTemplate):
    key = cuda.threadfence
    cases = [signature(types.none)]


@register
class Cuda_threadfence_block(ConcreteTemplate):
    key = cuda.threadfence_block
    cases = [signature(types.none)]


@register
class Cuda_threadfence_system(ConcreteTemplate):
    key = cuda.threadfence_system
    cases = [signature(types.none)]


@register
class Cuda_syncwarp(ConcreteTemplate):
    key = cuda.syncwarp
    cases = [signature(types.none), signature(types.none, types.i4)]


@register
class Cuda_match_any_sync(ConcreteTemplate):
    key = cuda.match_any_sync
    cases = [
        signature(types.i4, types.i4, types.i4),
        signature(types.i4, types.i4, types.i8),
        signature(types.i4, types.i4, types.f4),
        signature(types.i4, types.i4, types.f8),
    ]


@register
class Cuda_match_all_sync(ConcreteTemplate):
    key = cuda.match_all_sync
    cases = [
        signature(types.Tuple((types.i4, types.b1)), types.i4, types.i4),
        signature(types.Tuple((types.i4, types.b1)), types.i4, types.i8),
        signature(types.Tuple((types.i4, types.b1)), types.i4, types.f4),
        signature(types.Tuple((types.i4, types.b1)), types.i4, types.f8),
    ]


@register
class Cuda_activemask(ConcreteTemplate):
    key = cuda.activemask
    cases = [signature(types.uint32)]


@register
class Cuda_lanemask_lt(ConcreteTemplate):
    key = cuda.lanemask_lt
    cases = [signature(types.uint32)]


@register
class Cuda_popc(ConcreteTemplate):
    """
    Supported types from `llvm.popc`
    [here](http://docs.nvidia.com/cuda/nvvm-ir-spec/index.html#bit-manipulations-intrinics)
    """

    key = cuda.popc
    cases = [
        signature(types.int8, types.int8),
        signature(types.int16, types.int16),
        signature(types.int32, types.int32),
        signature(types.int64, types.int64),
        signature(types.uint8, types.uint8),
        signature(types.uint16, types.uint16),
        signature(types.uint32, types.uint32),
        signature(types.uint64, types.uint64),
    ]


@register
class Cuda_fma(ConcreteTemplate):
    """
    Supported types from `llvm.fma`
    [here](https://docs.nvidia.com/cuda/nvvm-ir-spec/index.html#standard-c-library-intrinics)
    """

    key = cuda.fma
    cases = [
        signature(types.float32, types.float32, types.float32, types.float32),
        signature(types.float64, types.float64, types.float64, types.float64),
    ]


@register
class Cuda_cbrt(ConcreteTemplate):
    key = cuda.cbrt
    cases = [
        signature(types.float32, types.float32),
        signature(types.float64, types.float64),
    ]


@register
class Cuda_brev(ConcreteTemplate):
    key = cuda.brev
    cases = [
        signature(types.uint32, types.uint32),
        signature(types.uint64, types.uint64),
    ]


@register
class Cuda_clz(ConcreteTemplate):
    """
    Supported types from `llvm.ctlz`
    [here](http://docs.nvidia.com/cuda/nvvm-ir-spec/index.html#bit-manipulations-intrinics)
    """

    key = cuda.clz
    cases = [
        signature(types.int8, types.int8),
        signature(types.int16, types.int16),
        signature(types.int32, types.int32),
        signature(types.int64, types.int64),
        signature(types.uint8, types.uint8),
        signature(types.uint16, types.uint16),
        signature(types.uint32, types.uint32),
        signature(types.uint64, types.uint64),
    ]


@register
class Cuda_ffs(ConcreteTemplate):
    """
    Supported types from `llvm.cttz`
    [here](http://docs.nvidia.com/cuda/nvvm-ir-spec/index.html#bit-manipulations-intrinics)
    """

    key = cuda.ffs
    cases = [
        signature(types.uint32, types.int8),
        signature(types.uint32, types.int16),
        signature(types.uint32, types.int32),
        signature(types.uint32, types.int64),
        signature(types.uint32, types.uint8),
        signature(types.uint32, types.uint16),
        signature(types.uint32, types.uint32),
        signature(types.uint32, types.uint64),
    ]


@register
class Cuda_selp(AbstractTemplate):
    key = cuda.selp

    def generic(self, args, kws):
        assert not kws
        test, a, b = args

        # per docs
        # http://docs.nvidia.com/cuda/parallel-thread-execution/index.html#comparison-and-selection-instructions-selp
        supported_types = (
            types.float64,
            types.float32,
            types.int16,
            types.uint16,
            types.int32,
            types.uint32,
            types.int64,
            types.uint64,
        )

        if a != b or a not in supported_types:
            return

        return signature(a, test, a, a)


@register_global(float)
class Float(AbstractTemplate):
    def generic(self, args, kws):
        assert not kws

        [arg] = args

        if arg == types.float16:
            return signature(arg, arg)


# If multiple ConcreteTemplates provide typing for a single function, then
# function resolution will pick the first compatible typing it finds even if it
# involves inserting a cast that would be considered undesirable (in this
# specific case, float16s could be cast to float32s for comparisons).
#
# To work around this, we instead use an AbstractTemplate that implements
# exactly the casting logic that we desire. The AbstractTemplate gets
# considered in preference to ConcreteTemplates during typing.
#
# This is tracked as Issue #7863 (https://github.com/numba/numba/issues/7863) -
# once this is resolved it should be possible to replace this AbstractTemplate
# with a ConcreteTemplate to simplify the logic.


# generate atomic operations
def _gen(l_key, supported_types):
    @register
    class Cuda_atomic(AbstractTemplate):
        key = l_key

        def generic(self, args, kws):
            assert not kws
            ary, idx, val = args

            if ary.dtype not in supported_types:
                return

            if ary.ndim == 1:
                return signature(ary.dtype, ary, types.intp, ary.dtype)
            elif ary.ndim > 1:
                return signature(ary.dtype, ary, idx, ary.dtype)

    return Cuda_atomic


all_numba_types = (
    types.float64,
    types.float32,
    types.int32,
    types.uint32,
    types.int64,
    types.uint64,
)

integer_numba_types = (types.int32, types.uint32, types.int64, types.uint64)

unsigned_int_numba_types = (types.uint32, types.uint64)

Cuda_atomic_add = _gen(cuda.atomic.add, all_numba_types)
Cuda_atomic_sub = _gen(cuda.atomic.sub, all_numba_types)
Cuda_atomic_max = _gen(cuda.atomic.max, all_numba_types)
Cuda_atomic_min = _gen(cuda.atomic.min, all_numba_types)
Cuda_atomic_nanmax = _gen(cuda.atomic.nanmax, all_numba_types)
Cuda_atomic_nanmin = _gen(cuda.atomic.nanmin, all_numba_types)
Cuda_atomic_and = _gen(cuda.atomic.and_, integer_numba_types)
Cuda_atomic_or = _gen(cuda.atomic.or_, integer_numba_types)
Cuda_atomic_xor = _gen(cuda.atomic.xor, integer_numba_types)
Cuda_atomic_inc = _gen(cuda.atomic.inc, unsigned_int_numba_types)
Cuda_atomic_dec = _gen(cuda.atomic.dec, unsigned_int_numba_types)
Cuda_atomic_exch = _gen(cuda.atomic.exch, integer_numba_types)


@register
class Cuda_atomic_compare_and_swap(AbstractTemplate):
    key = cuda.atomic.compare_and_swap

    def generic(self, args, kws):
        assert not kws
        ary, old, val = args
        dty = ary.dtype

        if dty in integer_numba_types and ary.ndim == 1:
            return signature(dty, ary, dty, dty)


@register
class Cuda_atomic_cas(AbstractTemplate):
    key = cuda.atomic.cas

    def generic(self, args, kws):
        assert not kws
        ary, idx, old, val = args
        dty = ary.dtype

        if dty not in integer_numba_types:
            return

        if ary.ndim == 1:
            return signature(dty, ary, types.intp, dty, dty)
        elif ary.ndim > 1:
            return signature(dty, ary, idx, dty, dty)


@register_global(breakpoint)
class Cuda_breakpoint(ConcreteTemplate):
    cases = [signature(types.none)]


@register
class Cuda_nanosleep(ConcreteTemplate):
    key = cuda.nanosleep

    cases = [signature(types.void, types.uint32)]


@register_attr
class Dim3_attrs(AttributeTemplate):
    key = dim3

    def resolve_x(self, mod):
        return types.int32

    def resolve_y(self, mod):
        return types.int32

    def resolve_z(self, mod):
        return types.int32


@register_attr
class CudaSharedModuleTemplate(AttributeTemplate):
    key = types.Module(cuda.shared)

    def resolve_array(self, mod):
        return types.Function(Cuda_shared_array)


@register_attr
class CudaConstModuleTemplate(AttributeTemplate):
    key = types.Module(cuda.const)

    def resolve_array_like(self, mod):
        return types.Function(Cuda_const_array_like)


@register_attr
class CudaLocalModuleTemplate(AttributeTemplate):
    key = types.Module(cuda.local)

    def resolve_array(self, mod):
        return types.Function(Cuda_local_array)


@register_attr
class CudaAtomicTemplate(AttributeTemplate):
    key = types.Module(cuda.atomic)

    def resolve_add(self, mod):
        return types.Function(Cuda_atomic_add)

    def resolve_sub(self, mod):
        return types.Function(Cuda_atomic_sub)

    def resolve_and_(self, mod):
        return types.Function(Cuda_atomic_and)

    def resolve_or_(self, mod):
        return types.Function(Cuda_atomic_or)

    def resolve_xor(self, mod):
        return types.Function(Cuda_atomic_xor)

    def resolve_inc(self, mod):
        return types.Function(Cuda_atomic_inc)

    def resolve_dec(self, mod):
        return types.Function(Cuda_atomic_dec)

    def resolve_exch(self, mod):
        return types.Function(Cuda_atomic_exch)

    def resolve_max(self, mod):
        return types.Function(Cuda_atomic_max)

    def resolve_min(self, mod):
        return types.Function(Cuda_atomic_min)

    def resolve_nanmin(self, mod):
        return types.Function(Cuda_atomic_nanmin)

    def resolve_nanmax(self, mod):
        return types.Function(Cuda_atomic_nanmax)

    def resolve_compare_and_swap(self, mod):
        return types.Function(Cuda_atomic_compare_and_swap)

    def resolve_cas(self, mod):
        return types.Function(Cuda_atomic_cas)


@register_attr
class CudaModuleTemplate(AttributeTemplate):
    key = types.Module(cuda)

    def resolve_cg(self, mod):
        return types.Module(cuda.cg)

    def resolve_threadIdx(self, mod):
        return dim3

    def resolve_blockIdx(self, mod):
        return dim3

    def resolve_blockDim(self, mod):
        return dim3

    def resolve_gridDim(self, mod):
        return dim3

    def resolve_laneid(self, mod):
        return types.int32

    def resolve_shared(self, mod):
        return types.Module(cuda.shared)

    def resolve_popc(self, mod):
        return types.Function(Cuda_popc)

    def resolve_brev(self, mod):
        return types.Function(Cuda_brev)

    def resolve_clz(self, mod):
        return types.Function(Cuda_clz)

    def resolve_ffs(self, mod):
        return types.Function(Cuda_ffs)

    def resolve_fma(self, mod):
        return types.Function(Cuda_fma)

    def resolve_cbrt(self, mod):
        return types.Function(Cuda_cbrt)

    def resolve_threadfence(self, mod):
        return types.Function(Cuda_threadfence_device)

    def resolve_threadfence_block(self, mod):
        return types.Function(Cuda_threadfence_block)

    def resolve_threadfence_system(self, mod):
        return types.Function(Cuda_threadfence_system)

    def resolve_syncwarp(self, mod):
        return types.Function(Cuda_syncwarp)

    def resolve_match_any_sync(self, mod):
        return types.Function(Cuda_match_any_sync)

    def resolve_match_all_sync(self, mod):
        return types.Function(Cuda_match_all_sync)

    def resolve_activemask(self, mod):
        return types.Function(Cuda_activemask)

    def resolve_lanemask_lt(self, mod):
        return types.Function(Cuda_lanemask_lt)

    def resolve_selp(self, mod):
        return types.Function(Cuda_selp)

    def resolve_nanosleep(self, mod):
        return types.Function(Cuda_nanosleep)

    def resolve_atomic(self, mod):
        return types.Module(cuda.atomic)

    def resolve_const(self, mod):
        return types.Module(cuda.const)

    def resolve_local(self, mod):
        return types.Module(cuda.local)


register_global(cuda, types.Module(cuda))



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba.cuda.cudadrv.error import (
    CCSupportError,
)
from numba.cuda import config
from numba.cuda.cuda_paths import get_cuda_paths
from numba.cuda.utils import _readenv
from cuda import pathfinder
import os
import warnings
import functools

from cuda.core import Program, ProgramOptions
from cuda.bindings import nvrtc as bindings_nvrtc

NVRTC_EXTRA_SEARCH_PATHS = _readenv(
    "NUMBA_CUDA_NVRTC_EXTRA_SEARCH_PATHS", str, ""
) or getattr(config, "CUDA_NVRTC_EXTRA_SEARCH_PATHS", "")
if not hasattr(config, "CUDA_NVRTC_EXTRA_SEARCH_PATHS"):
    config.CUDA_NVRTC_EXTRA_SEARCH_PATHS = NVRTC_EXTRA_SEARCH_PATHS


@functools.cache
def _get_nvrtc_version():
    retcode, major, minor = bindings_nvrtc.nvrtcVersion()
    if retcode != bindings_nvrtc.nvrtcResult.NVRTC_SUCCESS:
        raise RuntimeError(f"{retcode.name} when calling nvrtcVersion()")
    return (major, minor)


def _verify_cc_tuple(cc):
    version = _get_nvrtc_version()
    ver_str = lambda version: ".".join(str(v) for v in version)

    if len(cc) == 3:
        cc, arch = (cc[0], cc[1]), cc[2]
    else:
        arch = ""

    if arch not in ("", "a", "f"):
        raise ValueError(
            f"Invalid architecture suffix '{arch}' in compute capability "
            f"{ver_str(cc)}{arch}. Expected '', 'a', or 'f'."
        )

    supported_ccs = get_supported_ccs()
    try:
        found = max(filter(lambda v: v <= cc, [v for v in supported_ccs]))
    except ValueError:
        raise RuntimeError(
            f"Device compute capability {ver_str(cc)} is less than the "
            f"minimum supported by NVRTC {ver_str(version)}. Supported "
            "compute capabilities are "
            f"{', '.join([ver_str(v) for v in supported_ccs])}."
        )

    if found != cc:
        found = (found[0], found[1], arch)
        warnings.warn(
            f"Device compute capability {ver_str(cc)} is not supported by "
            f"NVRTC {ver_str(version)}. Using {ver_str(found)} instead."
        )
    else:
        found = (cc[0], cc[1], arch)

    return found


def compile(src, name, cc, ltoir=False, lineinfo=False, debug=False):
    """
    Compile a CUDA C/C++ source to PTX or LTOIR for a given compute capability.

    :param src: The source code to compile
    :type src: str
    :param name: The filename of the source (for information only)
    :type name: str
    :param cc: A tuple ``(major, minor)`` or ``(major, minor, arch)`` of the
        compute capability
    :type cc: tuple
    :param ltoir: Compile into LTOIR if True, otherwise into PTX
    :type ltoir: bool
    :param lineinfo: Whether to include line information in the compiled code
    :type lineinfo: bool
    :param debug: Whether to include debug information in the compiled code
    :type debug: bool
    :return: The compiled PTX or LTOIR and compilation log
    :rtype: tuple
    """
    found = _verify_cc_tuple(cc)
    version = _get_nvrtc_version()

    # Compilation options:
    # - Compile for the current device's compute capability.
    # - The CUDA include path is added.
    # - Relocatable Device Code (rdc) is needed to prevent device functions
    #   being optimized away.
    major, minor = found[0], found[1]
    cc_arch = found[2] if len(found) == 3 else ""

    arch = f"sm_{major}{minor}{cc_arch}"

    cuda_include_dir = get_cuda_paths()["include_dir"].info
    cuda_includes = [f"{cuda_include_dir}"]

    cudadrv_path = os.path.dirname(os.path.abspath(__file__))
    numba_cuda_path = os.path.dirname(cudadrv_path)

    nvrtc_ver_major = version[0]

    if nvrtc_ver_major == 12:
        numba_include = f"{os.path.join(numba_cuda_path, 'include', '12')}"

    elif nvrtc_ver_major == 13:
        numba_include = f"{os.path.join(numba_cuda_path, 'include', '13')}"

    cccl_found_header_dir = pathfinder.locate_nvidia_header_directory("cccl")
    if cccl_found_header_dir is not None:
        # TODO: Not every kernel needs cccl, so it shouldn't
        # be added to the include path for every kernel.
        # Needs to be flagged during compilation if NRT is included.
        cccl_include_dir = cccl_found_header_dir.abs_path
        cuda_includes.append(cccl_include_dir)

    if config.CUDA_NVRTC_EXTRA_SEARCH_PATHS:
        extra_includes = config.CUDA_NVRTC_EXTRA_SEARCH_PATHS.split(":")
    else:
        extra_includes = []

    nrt_include = os.path.join(numba_cuda_path, "memory_management")

    includes = [numba_include, *cuda_includes, nrt_include, *extra_includes]

    # TODO: move all this into Program/ProgramOptions
    # logsz = config.CUDA_LOG_SIZE
    #
    # jitinfo = bytearray(logsz)
    # jiterrors = bytearray(logsz)
    #
    # jit_option = binding.CUjit_option
    # options = {
    #     jit_option.CU_JIT_INFO_LOG_BUFFER: jitinfo,
    #     jit_option.CU_JIT_INFO_LOG_BUFFER_SIZE_BYTES: logsz,
    #     jit_option.CU_JIT_ERROR_LOG_BUFFER: jiterrors,
    #     jit_option.CU_JIT_ERROR_LOG_BUFFER_SIZE_BYTES: logsz,
    #     jit_option.CU_JIT_LOG_VERBOSE: config.CUDA_VERBOSE_JIT_LOG,
    # }
    # info_log = jitinfo.decode("utf-8")

    options = ProgramOptions(
        arch=arch,
        include_path=includes,
        relocatable_device_code=True,
        link_time_optimization=ltoir,
        name=name,
        debug=debug,
        lineinfo=lineinfo,
    )

    class Logger:
        def __init__(self):
            self.log = []

        def write(self, msg):
            self.log.append(msg)

    logger = Logger()
    if isinstance(src, bytes):
        src = src.decode("utf8")

    prog = Program(src, "c++", options=options)
    result = prog.compile("ltoir" if ltoir else "ptx", logs=logger)
    log = ""
    if logger.log:
        log = logger.log
        joined_logs = "\n".join(log)
        warnings.warn(f"NVRTC log messages: {joined_logs}")
    return result, log


def find_closest_arch(cc):
    """
    Given a compute capability, return the closest compute capability supported
    by the CUDA toolkit.

    :param mycc: Compute capability as a tuple ``(MAJOR, MINOR)``
    :return: Closest supported CC as a tuple ``(MAJOR, MINOR)``
    """
    supported_ccs = get_supported_ccs()

    for i, supported_cc in enumerate(supported_ccs):
        if supported_cc == cc:
            # Matches
            return supported_cc
        elif supported_cc > cc:
            # Exceeded
            if i == 0:
                # CC lower than supported
                msg = (
                    "GPU compute capability %d.%d is not supported"
                    "(requires >=%d.%d)" % (cc + supported_cc)
                )
                raise CCSupportError(msg)
            else:
                # return the previous CC
                return supported_ccs[i - 1]

    # CC higher than supported
    return supported_ccs[-1]  # Choose the highest


def get_arch_option(major, minor, arch=""):
    """Matches with the closest architecture option"""
    if config.FORCE_CUDA_CC:
        fcc = config.FORCE_CUDA_CC
        major, minor = fcc[0], fcc[1]
        if len(fcc) == 3:
            arch = fcc[2]
        else:
            arch = ""
    else:
        new_major, new_minor = find_closest_arch((major, minor))
        if (new_major, new_minor) != (major, minor):
            # If we picked a different major / minor, then using an
            # arch-specific version is invalid
            if arch != "":
                raise ValueError(
                    f"Can't use arch-specific compute_{major}{minor}{arch} with "
                    "closest found compute capability "
                    f"compute_{new_major}{new_minor}"
                )
        major, minor = new_major, new_minor

    return f"compute_{major}{minor}{arch}"


def get_lowest_supported_cc():
    return min(get_supported_ccs())


def get_supported_ccs():
    retcode, archs = bindings_nvrtc.nvrtcGetSupportedArchs()
    if retcode != bindings_nvrtc.nvrtcResult.NVRTC_SUCCESS:
        raise RuntimeError(
            f"{retcode.name} when calling nvrtcGetSupportedArchs()"
        )
    return [(arch // 10, arch % 10) for arch in archs]
